#include "ctalk.h"
#include "compiler.h"
#include <time.h>
#include <ctype.h>
#ifdef _WIN32
#include <direct.h>
#endif

static CtkObject lockMutex(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    CHECK_PRIMITIVE_ARGUMENT(args[0], CTK_MUTEX);
    CtkThread* thr = ctkGetCurrentThread();
    CtkMemoryManager::instance.endAccess(thr);
    ctkLockMutex((CtkMutex*)args[0].u.ptr);
    CtkMemoryManager::instance.beginAccess(thr);
    return ctkNull;
}

static CtkPrimitive p1("lockMutex", lockMutex);

static CtkObject unlockMutex(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    CHECK_PRIMITIVE_ARGUMENT(args[0], CTK_MUTEX);
    ctkUnlockMutex((CtkMutex*)args[0].u.ptr);
    return ctkNull;
}
static CtkPrimitive p2("unlockMutex", unlockMutex);

static CtkObject createMutex(int nArgs, CtkObject*) {
    CHECK_PRIMITIVE_N_ARGS(0, nArgs);
    CtkObject result;
    MAKE_MUTEX(result, ctkCreateMutex());
    return result;
}
static CtkPrimitive p3("createMutex", createMutex);

static CtkObject deleteMutex(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    CHECK_PRIMITIVE_ARGUMENT(args[0], CTK_MUTEX);
    ctkDeleteMutex((CtkMutex*)args[0].u.ptr);
    return ctkNull;
}
static CtkPrimitive p4("deleteMutex", deleteMutex);


CtkObject deleteArray(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    CHECK_PRIMITIVE_ARGUMENT(args[0], CTK_ARRAY);
    ctkDeleteArray((CtkArray*)args[0].u.ptr);
    return ctkNull;
}
static CtkPrimitive p5("deleteArray", deleteArray);


CtkObject clearArray(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    CHECK_PRIMITIVE_ARGUMENT(args[0], CTK_ARRAY);
    ctkClearArray((CtkArray*)args[0].u.ptr);
    return ctkNull;
}
static CtkPrimitive p6("clearArray", clearArray);

CtkObject cloneArray(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    CHECK_PRIMITIVE_ARGUMENT(args[0], CTK_ARRAY);
    CtkObject result;
    MAKE_ARRAY(result, ctkCloneArray((CtkArray*)args[0].u.ptr));
    return result;
}
static CtkPrimitive p7("cloneArray", cloneArray);


CtkObject getKeys(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_ARGUMENT(args[0], CTK_ARRAY);
    CtkObject result;
    CtkArray* dst = ctkCreateArray();
    CtkArray* src = (CtkArray*)args[0].u.ptr; 
    CtkObject key;
    key.type = CTK_INTEGER;
    for (int i = src->nAllocated, j = 0; --i >= 0;) { 
	for (CtkHashEntry* entry = src->entries[i]; entry != NULL; entry = entry->next) { 
	    key.u.ivalue = j++;
	    ctkPutArray(dst, key, entry->key);
	}
    }
    MAKE_ARRAY(result, dst);
    return result;
}
static CtkPrimitive p77("getKeys", getKeys);

static CtkObject getMessage(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    CHECK_PRIMITIVE_ARGUMENT(args[0], CTK_THREAD);
    CtkObject object = ctkGetMessage((CtkThread*)args[0].u.ptr);
    return object;
}
static CtkPrimitive p8("getMessage", getMessage);

static CtkObject putMessage(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(2, nArgs);
    CHECK_PRIMITIVE_ARGUMENT(args[0], CTK_THREAD);
    ctkPutMessage((CtkThread*)args[0].u.ptr, args[1]);
    return args[0];
}
static CtkPrimitive p9("putMessage", putMessage);

CtkObject currentThread(int nArgs, CtkObject*) {
    CHECK_PRIMITIVE_N_ARGS(0, nArgs);
    CtkObject result;
    MAKE_THREAD(result, ctkGetCurrentThread());
    return result;
}
static CtkPrimitive p10("currentThread", currentThread);

static CtkObject createFile(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(2, nArgs);
    CHECK_PRIMITIVE_ARGUMENT(args[0], CTK_STRING);
    CHECK_PRIMITIVE_ARGUMENT(args[1], CTK_STRING);
    FILE* f = fopen(args[0].u.svalue, args[1].u.svalue);
    if (f == NULL) { 
	return ctkNull;
    }
    CtkObject result;
    MAKE_FILE(result, ctkCreateFile(args[0].u.svalue, args[1].u.svalue));
    return result;
}
static CtkPrimitive p11("createFile", createFile);

static CtkObject deleteFile(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    CHECK_PRIMITIVE_ARGUMENT(args[0], CTK_FILE);
    ctkDeleteFile((CtkFile*)args[0].u.ptr);
    return ctkNull;
}
static CtkPrimitive p12("deleteFile", deleteFile);

static CtkObject flushFile(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    CHECK_PRIMITIVE_ARGUMENT(args[0], CTK_FILE);
    ctkFlushFile((CtkFile*)args[0].u.ptr);
    return ctkNull;
}
static CtkPrimitive p13("flushFile", flushFile);


static CtkObject printFile(int nArgs, CtkObject* args) {
    if (nArgs < 1 || args[0].type != CTK_FILE) { 
	ctkThrowException("No file specified");
    }
    for (int i = 1; i < nArgs; i++) { 
	ctkPrintFile((CtkFile*)args[0].u.ptr, args[i]);
    }
    return args[0];
}
static CtkPrimitive p14("printFile", printFile);

static CtkObject printlnFile(int nArgs, CtkObject* args) {
    printFile(nArgs, args);
    CtkObject nl;
    MAKE_STRING(nl, "\n");
    ctkPrintFile((CtkFile*)args[0].u.ptr, nl);
    return args[0];
}
static CtkPrimitive p15("printlnFile", printlnFile);

static CtkObject print(int nArgs, CtkObject* args) {
    for (int i = 0; i < nArgs; i++) { 
	ctkPrint(args[i]);
    }
    return ctkNull;
}
static CtkPrimitive p16("print", print);

static CtkObject println(int nArgs, CtkObject* args) {
    print(nArgs, args);
    CtkObject nl;
    MAKE_STRING(nl, "\n");
    ctkPrint(nl);
    return ctkNull;
}
static CtkPrimitive p17("println", println);

#ifdef __sun__
#include <errno.h>
#include <sys/wait.h>

static int my_system (const char *command) {
    int pid, status;
    
    if (command == 0)
	return 1;
    pid = fork();
    if (pid == -1)
	return -1;
    if (pid == 0) {
	char *argv[4];
	argv[0] = "sh";
	argv[1] = "-c";
	argv[2] = (char*)command;
	argv[3] = 0;
	execv("/bin/sh", argv);
	exit(127);
    }
    do {
	if (waitpid(pid, &status, 0) == -1) {
	    if (errno != EINTR)
		return -1;
	} else
	    return status;
    } while(1);
}
#define system my_system
#endif

static CtkObject execute(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    CHECK_PRIMITIVE_ARGUMENT(args[0], CTK_STRING);
    CtkObject result;
    CtkThread* thr = ctkGetCurrentThread();
    CtkMemoryManager::instance.endAccess(thr);
    MAKE_INTEGER(result, system(args[0].u.svalue));
    CtkMemoryManager::instance.beginAccess(thr);
    return result;
}
static CtkPrimitive p18("execute", execute);

static CtkObject currentTime(int nArgs, CtkObject*) {
    CHECK_PRIMITIVE_N_ARGS(0, nArgs);
    CtkObject result;
    MAKE_INTEGER(result, time(NULL));
    return result;
}
static CtkPrimitive p19("time", currentTime);

static CtkObject currentTimeMillis(int nArgs, CtkObject*) {
    CHECK_PRIMITIVE_N_ARGS(0, nArgs);
    CtkObject result;
    MAKE_INTEGER(result, ctkGetTimeMillis());
    return result;
}
static CtkPrimitive p19a("timeMillis", currentTimeMillis);

static CtkObject assertFunc(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    if (!IS_TRUE(args[0])) { 
	ctkThrowException("Assertion failed");
    }
    return ctkNull;
} 
static CtkPrimitive p20("assert", assertFunc);

static CtkObject loadDLL(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    CHECK_PRIMITIVE_ARGUMENT(args[0], CTK_STRING);
    return ctkLoadLibrary(args[0].u.svalue) ? ctkTrue : ctkFalse;
} 
static CtkPrimitive p21("loadDLL", loadDLL);

static CtkObject printExceptionStackTrace(int nArgs, CtkObject*) {
    CHECK_PRIMITIVE_N_ARGS(0, nArgs);
    CtkThread* thread = ctkGetCurrentThread();
    if (thread->exceptionTrace != NULL) { 
	ctkPrintStackTrace(thread->exceptionTrace);
    }
    return ctkNull;
}
static CtkPrimitive p22("printExceptionStackTrace", printExceptionStackTrace);



static CtkObject loadModule(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    CHECK_PRIMITIVE_ARGUMENT(args[0], CTK_STRING);
    return CtkCompiler::instance.compile(args[0].u.svalue) != NULL ? ctkTrue : ctkFalse;
} 
static CtkPrimitive p23("loadModule", loadModule);


static CtkObject indexOf(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(2, nArgs);
    CHECK_PRIMITIVE_ARGUMENT(args[0], CTK_STRING);
    CHECK_PRIMITIVE_ARGUMENT(args[1], CTK_STRING);
    char* s1 = args[0].u.svalue;
    char* s2 = args[1].u.svalue;
    int n1 = strlen(s1);
    int n2 = strlen(s2);
    CtkObject result;
    int i, j;
    for (i = 0; i <= n1-n2; i++) { 
	for (j = 0; j < n2 && s1[i+j] == s2[j]; j++);
	if (j == n2) { 
	    MAKE_INTEGER(result, i);
	    return result;
	}
    }
    MAKE_INTEGER(result, -1);
    return result;
}    
static CtkPrimitive p24("indexOf", indexOf);


static CtkObject lastIndexOf(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(2, nArgs);
    CHECK_PRIMITIVE_ARGUMENT(args[0], CTK_STRING);
    CHECK_PRIMITIVE_ARGUMENT(args[1], CTK_STRING);
    char* s1 = args[0].u.svalue;
    char* s2 = args[1].u.svalue;
    int n1 = strlen(s1);
    int n2 = strlen(s2);
    CtkObject result;
    int i, j;
    for (i = n1-n2; i >= 0; i--) { 
	for (j = 0; j < n2 && s1[i+j] == s2[j]; j++);
	if (j == n2) { 
	    MAKE_INTEGER(result, i);
	    return result;
	}
    }
    MAKE_INTEGER(result, -1);
    return result;
}    
static CtkPrimitive p25("lastIndexOf", lastIndexOf);

static CtkObject substring(int nArgs, CtkObject* args) {
    if (nArgs < 2 || nArgs > 3) { 
	ctkThrowException("Wrong number of primitive arguments");
    }
    CHECK_PRIMITIVE_ARGUMENT(args[0], CTK_STRING);
    char* src = args[0].u.svalue;
    unsigned len = strlen(src);
    CHECK_PRIMITIVE_ARGUMENT(args[1], CTK_INTEGER);
    unsigned from = args[1].u.ivalue;
    if (from > len) { 
	from = len;
    }
    unsigned till = len;
    if (nArgs == 3) { 
	CHECK_PRIMITIVE_ARGUMENT(args[2], CTK_INTEGER);
	till = args[2].u.ivalue;
	if (till < from) { 
	    till = from;
	} else if (till > len) { 
	    till = len;
	}
    }
    char* dst = (char*)ctkAllocateObject(till-from+1, NULL);
    memcpy(dst, src + from, till-from);
    dst[till-from] = '\0';
    CtkObject result;
    MAKE_STRING(result, dst);
    return result;
}    
static CtkPrimitive p26("substring", substring);

static CtkObject toLower(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    CHECK_PRIMITIVE_ARGUMENT(args[0], CTK_STRING);
    char* src = args[0].u.svalue;
    char* dst = (char*)ctkAllocateObject(strlen(src)+1, NULL);
    CtkObject result;
    MAKE_STRING(result, dst);
    while ((*dst++ = tolower((unsigned char)*src++)) != 0);
    return result;
}
static CtkPrimitive p27("toLower", toLower);

static CtkObject toUpper(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    CHECK_PRIMITIVE_ARGUMENT(args[0], CTK_STRING);
    char* src = args[0].u.svalue;
    char* dst = (char*)ctkAllocateObject(strlen(src)+1, NULL);
    CtkObject result;
    MAKE_STRING(result, dst);
    while ((*dst++ = toupper((unsigned char)*src++)) != 0);
    return result;
}
static CtkPrimitive p28("toUpper", toUpper);

static CtkObject trim(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    CHECK_PRIMITIVE_ARGUMENT(args[0], CTK_STRING);
    char* src = args[0].u.svalue;
    int start = 0, end = strlen(src);
    while ((unsigned char)(src[start]-1) <= ' ') { 
        start += 1;
    }
    while (end > start && (unsigned char)src[end-1] <= ' ') { 
	end -= 1;
    }
    char* dst = (char*)ctkAllocateObject(end-start+1, NULL);
    CtkObject result;
    MAKE_STRING(result, dst);
    memcpy(dst, src+start, end-start);
    dst[end-start] = '\0';
    return result;
}
static CtkPrimitive p29("trim", trim);

static CtkObject startsWith(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(2, nArgs);
    CHECK_PRIMITIVE_ARGUMENT(args[0], CTK_STRING);
    CHECK_PRIMITIVE_ARGUMENT(args[1], CTK_STRING);
    char* p = args[0].u.svalue;
    char* q = args[1].u.svalue;
    while (*q != '\0') { 
	if (*p != *q) { 
	    return ctkFalse;
	}
    }
    return ctkTrue;
}
static CtkPrimitive p30("startsWith", startsWith);

static CtkObject endsWith(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(2, nArgs);
    CHECK_PRIMITIVE_ARGUMENT(args[0], CTK_STRING);
    CHECK_PRIMITIVE_ARGUMENT(args[1], CTK_STRING);
    char* s1 = args[0].u.svalue;
    char* s2 = args[1].u.svalue;
    int n1 = strlen(s1);
    int n2 = strlen(s2);
    return (n1 >= n2 && strcmp(s1 + n1 - n2, s2) == 0) ? ctkTrue : ctkFalse;
}
static CtkPrimitive p31("endsWith", endsWith);

static CtkObject string(int nArgs, CtkObject* args) {
    char buf[256];
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    if (args[0].type == CTK_STRING) { 
	return args[0];
    }
    ctkToString(args[0], buf);
    char* str = (char*)ctkAllocateObject(strlen(buf) + 1, NULL);
    strcpy(str, buf);
    CtkObject result;
    MAKE_STRING(result, str);
    return result;
}
static CtkPrimitive p32("string", string);

static CtkObject length(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    int len = 0;
    switch (args[0].type) { 
      case CTK_STRING:
	len = strlen(args[0].u.svalue);
	break;
      case CTK_ARRAY:
	len = ((CtkArray*)args[0].u.ptr)->nUsed;
	break;
      default:
	ctkThrowException("Incompatible type of primitive parameter");
    } 
    CtkObject result;
    MAKE_INTEGER(result, len);
    return result;
}
static CtkPrimitive p33("length", length);

static CtkObject chDir(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    CHECK_PRIMITIVE_ARGUMENT(args[0], CTK_STRING);
    return chdir(args[0].u.svalue) == 0 ? ctkTrue : ctkFalse;
}
static CtkPrimitive p34("chdir", chDir);

static CtkObject pwd(int nArgs, CtkObject*) {
    CHECK_PRIMITIVE_N_ARGS(0, nArgs);
    char buf[1024];    
    if (getcwd(buf, sizeof buf) == NULL) {
	return ctkNull;
    }
    CtkObject result;
    ALLOCATE_STRING(result, buf);
    return result;
}
static CtkPrimitive p35("pwd", pwd);

static CtkObject sleepSec(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    CHECK_PRIMITIVE_ARGUMENT(args[0], CTK_INTEGER);
    CtkThread* thr = ctkGetCurrentThread();
    CtkMemoryManager::instance.endAccess(thr);
    ctkSleep(args[0].u.ivalue);
    CtkMemoryManager::instance.beginAccess(thr);
    return ctkNull;
}
static CtkPrimitive p36("sleep", sleepSec);

static CtkObject input(int nArgs, CtkObject* args) {
    char buf[1024];
    FILE* f = stdin;
    if (nArgs == 1) { 
	switch (args[0].type) { 
	  case CTK_STRING:
	    fputs(args[0].u.svalue, stdout);
	    break;
	  case CTK_FILE:
	    f = ((CtkFile*)args[0].u.ptr)->f;
	    break;
	  default: 
	    ctkThrowException("Incompatible type of primitive parameter");
	}
    } else { 
	CHECK_PRIMITIVE_N_ARGS(0, nArgs);
    }
    CtkObject result = ctkNull;
    CtkThread* thr = ctkGetCurrentThread();
    CtkMemoryManager::instance.endAccess(thr);
    if (fgets(buf, sizeof buf, f)) { 
	ALLOCATE_STRING(result, buf);
    }
    CtkMemoryManager::instance.beginAccess(thr);
    return result;
}
static CtkPrimitive p37("input", input);

static CtkObject integer(int nArgs, CtkObject* args) {
    CtkObject result;
    char ch;

    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    switch (args[0].type) { 
      case CTK_INTEGER:
	return args[0];
      case CTK_REAL:
	MAKE_INTEGER(result, (ctk_integer)args[0].u.rvalue);
	break;
      case CTK_STRING:
	if (sscanf(args[0].u.svalue, "%ld%c", &result.u.ivalue, &ch) != 1) { 
	    ctkThrowException("Invalid number format");
	}
	result.type = CTK_INTEGER;
	break;
      case CTK_RAW_POINTER:
	MAKE_INTEGER(result, (ctk_integer)args[0].u.ptr);
	break;        
      default: 
	ctkThrowException("Incompatible type of primitive parameter");
    }
    return result;
}
static CtkPrimitive p38("integer", integer);

static CtkObject real(int nArgs, CtkObject* args) {
    CtkObject result;
    char ch;

    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    switch (args[0].type) { 
      case CTK_INTEGER:
	MAKE_REAL(result, args[0].u.ivalue);
	break;	
      case CTK_REAL:
	return args[0];
      case CTK_STRING:
	if (sscanf(args[0].u.svalue, "%lf%c", &result.u.rvalue, &ch) != 1) { 
	    ctkThrowException("Invalid number format");
	}
	result.type = CTK_REAL;
	break;
      default: 
	ctkThrowException("Incompatible type of primitive parameter");
    }
    return result;
}
static CtkPrimitive p38r("real", real);

static CtkObject options(int nArgs, CtkObject* args) {
    if (nArgs == 1) { 
	return ctkGetArray(ctkGetCommandLineOptions(), args[0]);
    } else { 
	CtkObject result;
	CHECK_PRIMITIVE_N_ARGS(0, nArgs);
	MAKE_ARRAY(result, ctkGetCommandLineOptions());
	return result;
    }
}
static CtkPrimitive p39("options", options);


static CtkObject isInteger(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    return args->type == CTK_INTEGER ? ctkTrue : ctkFalse;
}
static CtkPrimitive p40("isInteger", isInteger);

static CtkObject isReal(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    return args->type == CTK_REAL ? ctkTrue : ctkFalse;
}
static CtkPrimitive p40r("isReal", isReal);


static CtkObject isString(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    return args->type == CTK_STRING ? ctkTrue : ctkFalse;
}
static CtkPrimitive p41("isString", isString);



static CtkObject isArray(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    return args->type == CTK_ARRAY ? ctkTrue : ctkFalse;
}
static CtkPrimitive p42("isArray", isArray);

static CtkObject isFile(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    return args->type == CTK_FILE ? ctkTrue : ctkFalse;
}
static CtkPrimitive p43("isFile", isFile);

static CtkObject isThread(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    return args->type == CTK_THREAD ? ctkTrue : ctkFalse;
}
static CtkPrimitive p44("isThread", isThread);



static CtkObject isNull(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    return args->type == CTK_NULL ? ctkTrue : ctkFalse;
}
static CtkPrimitive p45("isNull", isNull);


static CtkObject isMutex(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    return args->type == CTK_MUTEX ? ctkTrue : ctkFalse;
}
static CtkPrimitive p46("isMutex", isMutex);


static CtkObject isFunction(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    return args->type == CTK_FUNCTION ? ctkTrue : ctkFalse;
}
static CtkPrimitive p47("isFunction", isFunction);

static CtkObject isPrimitive(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    return args->type == CTK_PRIMITIVE ? ctkTrue : ctkFalse;
}
static CtkPrimitive p48("isPrimitive", isPrimitive);


static CtkObject getTypeName(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    static char const* typeMnem[] = { 
	"null", "integer", "real", "primitive", "raw_pointer", 
	"string", "array", "file", "mutex", "thread", "function", "frame", "user_type"
    };
    CtkObject result;
    ALLOCATE_STRING(result, typeMnem[args->type]);
    return result;
}
static CtkPrimitive p49("getTypeName", getTypeName);


static CtkObject getType(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    CtkObject result;
    MAKE_INTEGER(result, args->type);
    return result;
}
static CtkPrimitive p50("getType", getType);




