#include "ctalk.h"
#include "compiler.h"
#include <math.h>
#include <time.h>
#include <ctype.h>
#ifdef _WIN32
#include <direct.h>
#endif


#pragma warning(once: 4244)

void invalid_input()
{
	ctkThrowException(" Invalid arguments. ");
}

static CtkObject lockMutex(int nArgs, CtkObject* args) {
	if (nArgs < 1 || args[0].type != CTK_MUTEX) invalid_input();
    CtkThread* thr = ctkGetCurrentThread();
    CtkMemoryManager::instance.endAccess(thr);
    ctkLockMutex((CtkMutex*)args[0].u.ptr);
    CtkMemoryManager::instance.beginAccess(thr);
    return ctkNull;
}
static CtkPrimitive p0("lock", lockMutex);


static CtkObject unlockMutex(int nArgs, CtkObject* args) {
	if (nArgs < 1 || args[0].type != CTK_MUTEX) invalid_input();
    ctkUnlockMutex((CtkMutex*)args[0].u.ptr);
    return ctkNull;
}
static CtkPrimitive p1("unlock", unlockMutex);


static CtkObject ctkCreate(int nArgs, CtkObject *args) {
	if (nArgs < 0 || !IS_STRING(args[0])) return ctkNull;
    CtkObject ret;
	if (strcmp("mutex", args[0].u.svalue) == 0) {
		MAKE_MUTEX(ret, ctkCreateMutex());
		return ret;
	}
	if (strcmp("array", args[0].u.svalue) == 0) {
	}
    return ctkNull;
}
static CtkPrimitive p2("create", ctkCreate);


static CtkObject ctkDelete(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
	switch (args[0].type) {
	case CTK_MUTEX:
	    ctkDeleteMutex((CtkMutex*)args[0].u.ptr);
		break;
	case CTK_ARRAY:
		ctkDeleteArray((CtkArray*)args[0].u.ptr);
		break;
	}
    return ctkNull;
}
static CtkPrimitive p3("delete", ctkDelete);


CtkObject clearArray(int nArgs, CtkObject* args) {
	if (nArgs < 1 || args[0].type != CTK_ARRAY) invalid_input();
    ctkClearArray((CtkArray*)args[0].u.ptr);
    return ctkNull;
}
static CtkPrimitive p10("clear", clearArray);


CtkObject cloneArray(int nArgs, CtkObject* args) {
	if (nArgs < 1 || args[0].type != CTK_ARRAY) invalid_input();
    CtkObject result;
    MAKE_ARRAY(result, ctkCloneArray((CtkArray*)args[0].u.ptr));
    return result;
}
static CtkPrimitive p11("clone", cloneArray);


CtkObject ctkForeach(int nArgs, CtkObject* args) {
	if (nArgs < 2 ||
		!IS_ARRAY(args[0]) ||
		!IS_STRING(args[1])) invalid_input();

    CtkArray* arr = (CtkArray*)args[0].u.ptr; 
	CtkFunction* func = ctkGetFunction(args[1].u.svalue);
	if (func == NULL) ctkThrowException(" Invalid callback name.");

	for (int i = arr->nAllocated, j = 0; --i >= 0;) {
		for (CtkHashEntry* entry = arr->entries[i]; entry != NULL; entry = entry->next) {
			CtkObject result, params[2];
			params[0] = entry->key;
			params[1] = entry->value;
			result = ctkCallFunction(func, params);
			if (result.type = CTK_NULL) return result;
		}
    }

    return ctkNull;
}
static CtkPrimitive p12("foreach", ctkForeach);


static CtkObject getMessage(int nArgs, CtkObject* args) {
	if (nArgs < 1 || args[0].type != CTK_THREAD) invalid_input();
    CtkObject object = ctkGetMessage((CtkThread*)args[0].u.ptr);
    return object;
}
static CtkPrimitive p20("getmessage", getMessage);


static CtkObject putMessage(int nArgs, CtkObject* args) {
	if (nArgs < 2 || args[0].type != CTK_THREAD) invalid_input();
    ctkPutMessage((CtkThread*)args[0].u.ptr, args[1]);
    return args[0];
}
static CtkPrimitive p21("putmessage", putMessage);


CtkObject currentThread(int nArgs, CtkObject*) {
    CtkObject result;
    MAKE_THREAD(result, ctkGetCurrentThread());
    return result;
}
static CtkPrimitive p22("getthread", currentThread);


static CtkObject createFile(int nArgs, CtkObject* args) {
	if (nArgs < 2 || !IS_STRING(args[0]) || !IS_STRING(args[1])) invalid_input();
	CtkFile *file = ctkCreateFile(args[0].u.svalue, args[1].u.svalue);
	if (file == NULL) return ctkNull;
    CtkObject result;
    MAKE_FILE(result, file);
    return result;
}
static CtkPrimitive p30("open", createFile);


static CtkObject deleteFile(int nArgs, CtkObject* args) {
	if (nArgs < 1 || args[0].type != CTK_FILE) invalid_input();
    ctkDeleteFile((CtkFile*)args[0].u.ptr);
    return ctkNull;
}
static CtkPrimitive p31("close", deleteFile);


static CtkObject flushFile(int nArgs, CtkObject* args) {
	if (nArgs < 1 || args[0].type != CTK_FILE) invalid_input();
    ctkFlushFile((CtkFile*)args[0].u.ptr);
    return ctkNull;
}
static CtkPrimitive p32("flush", flushFile);


static CtkObject fileEOF(int nArgs, CtkObject* args) {
	if (nArgs < 1 || args[0].type != CTK_FILE) invalid_input();
	FILE *f = ((CtkFile*)args[0].u.ptr)->f;
	if (feof(f)) return ctkTrue;
	return ctkFalse;
}
static CtkPrimitive p33("eof", fileEOF);


static CtkObject filePrint(int nArgs, CtkObject* args) {
	if (nArgs < 1) invalid_input();
	FILE *f = stdout;
	int i = 0;
	if (IS_FILE(args[0])) {
		f = ((CtkFile*)args[0].u.ptr)->f;
		i = 1;
	}
	for (; i < nArgs; i++) {
	    if (args[i].type == CTK_STRING) { 
			fputs(args[i].u.svalue, f);
		} else { 
			char buf[256];
			ctkToString(args[i], buf);
			fputs(buf, f);
		}
		fflush(stdout);
	}
	return ctkNull;
}
static CtkPrimitive p34("print", filePrint);


static void delete_bytes(void *ptr)
{
	delete[] (char*)ptr;
}

static CtkObject fileRead(int nArgs, CtkObject* args) {
	if (nArgs < 2 ||
		!IS_FILE(args[0]) ||
		!IS_INTEGER(args[1]) ||
		args[1].u.ivalue < 1) invalid_input();
	char *ptr = new char[args[1].u.ivalue];
	if (!ptr) ctkThrowException(" : not enough memory.");
    CtkObject* o = (CtkObject*)ctkAllocateObject(sizeof(CtkObject), CtkFinalizer(delete_bytes));
	o->type = CTK_RAW_POINTER;
	o->u.ptr = ptr;
	size_t n = fread(ptr, 1, args[1].u.ivalue, ((CtkFile*)args[0].u.ptr)->f);
	if (n != args[1].u.ivalue) return ctkNull;
	return ctkTrue;
}
static CtkPrimitive p35("read", fileRead);


static CtkObject fileWrite(int nArgs, CtkObject* args) {
	if (nArgs < 3 ||
		!IS_FILE(args[0]) ||
		!IS_RAW_POINTER(args[1]) ||
		!IS_INTEGER(args[2]) ||
		args[2].u.ivalue < 1) invalid_input();
	size_t n = fwrite(args[1].u.ptr, 1, args[2].u.ivalue, ((CtkFile*)args[0].u.ptr)->f);
	if (n != args[2].u.ivalue) return ctkNull;
	return ctkTrue;
}
static CtkPrimitive p36("write", fileWrite);


static CtkObject fileSeek(int nArgs, CtkObject* args) {
	if (nArgs < 3 ||
		!IS_FILE(args[0]) ||
		!IS_INTEGER(args[1]) ||
		!IS_STRING(args[2])) invalid_input();
	FILE *f = ((CtkFile*)args[0].u.ptr)->f;
	long offset = args[2].u.ivalue;
	char *s = args[2].u.svalue;
	if (strcmp(s, "set") == 0) {
		if (fseek(f, offset, SEEK_SET) != 0) return ctkFalse;
	}
	else if (strcmp(s, "cur") == 0) {
		if (fseek(f, offset, SEEK_CUR) != 0) return ctkFalse;;
	}
	if (strcmp(s, "end") == 0) {
		if (fseek(f, offset, SEEK_END) != 0) return ctkFalse;;
	}
	return ctkTrue;
}
static CtkPrimitive p37("seek", fileSeek);


static CtkObject fileTell(int nArgs, CtkObject* args) {
	if (nArgs < 1 || !IS_FILE(args[0])) invalid_input();
	int pos = ftell(((CtkFile*)args[0].u.ptr)->f);
	CtkObject ret;
	MAKE_INTEGER(ret, pos);
	return ret;
}
static CtkPrimitive p38("tell", fileTell);


static CtkObject fileTmpnam(int nArgs, CtkObject* args) {
	CtkObject ret;
	MAKE_STRING(ret, tmpnam(0));
	return ret;
}
static CtkPrimitive p39("tmpname", fileTmpnam);


static CtkObject fileReadlines(int nArgs, CtkObject* args) {
	if (nArgs < 2 ||
		!IS_FILE(args[0]) ||
		!IS_STRING(args[1])) invalid_input();
	FILE *f = ((CtkFile*)args[0].u.ptr)->f;
	CtkFunction* func = ctkGetFunction(args[1].u.svalue);
	if (func == NULL) ctkThrowException(" Invalid callback name.");
	char buf[1025];
	while (!feof(f)) {
		if (fgets(buf, 1024, f)) {
			CtkObject result, params[1];
			MAKE_STRING(params[0], buf);
			result = ctkCallFunction(func, params);
			if (result.type = CTK_NULL) return result;
		}
	}
	return ctkNull;
}
static CtkPrimitive p40("readlines", fileReadlines);


static CtkObject input(int nArgs, CtkObject* args) {
    char buf[1024];
    FILE* f = stdin;
    if (nArgs == 1) { 
		switch (args[0].type) { 
		  case CTK_STRING:
		    fputs(args[0].u.svalue, stdout);
		    break;
		case CTK_FILE:
		    f = ((CtkFile*)args[0].u.ptr)->f;
		    break;
		default: 
		    ctkThrowException("Incompatible type of primitive parameter");
		}
    } else { 
		CHECK_PRIMITIVE_N_ARGS(0, nArgs);
    }
    CtkObject result = ctkNull;
    CtkThread* thr = ctkGetCurrentThread();
    CtkMemoryManager::instance.endAccess(thr);
    if (fgets(buf, sizeof buf, f)) {
		ALLOCATE_STRING(result, buf);
    }
    CtkMemoryManager::instance.beginAccess(thr);
    return result;
}
static CtkPrimitive p41("input", input);


#ifdef __sun__
#include <errno.h>
#include <sys/wait.h>

static int my_system (const char *command) {
    int pid, status;
    
    if (command == 0)
	return 1;
    pid = fork();
    if (pid == -1)
	return -1;
    if (pid == 0) {
	char *argv[4];
	argv[0] = "sh";
	argv[1] = "-c";
	argv[2] = (char*)command;
	argv[3] = 0;
	execv("/bin/sh", argv);
	exit(127);
    }
    do {
	if (waitpid(pid, &status, 0) == -1) {
	    if (errno != EINTR)
		return -1;
	} else
	    return status;
    } while(1);
}
#define system my_system
#endif

static CtkObject execute(int nArgs, CtkObject* args) {
	if (nArgs < 1 || !IS_STRING(args[0])) invalid_input();
    CtkObject result;
    CtkThread* thr = ctkGetCurrentThread();
    CtkMemoryManager::instance.endAccess(thr);
    MAKE_INTEGER(result, system(args[0].u.svalue));
    CtkMemoryManager::instance.beginAccess(thr);
    return result;
}
static CtkPrimitive p50("execute", execute);


static CtkObject currentTime(int nArgs, CtkObject*) {
	tm *st;
	time_t t = time(0);
	if (nArgs == 0) {
		st = localtime(&t);
	}
	else {
		st = gmtime(&t);
	} 
	CtkObject key, value;
	CtkArray* arr = ctkCreateArray();

	MAKE_STRING(key, "year");
	MAKE_INTEGER(value, st->tm_year);
	ctkPutArray(arr, key, value);
	
	MAKE_STRING(key, "month");
	MAKE_INTEGER(value, st->tm_mon);
	ctkPutArray(arr, key, value);
	
	MAKE_STRING(key, "day");
	MAKE_INTEGER(value, st->tm_mday);
	ctkPutArray(arr, key, value);
	
	MAKE_STRING(key, "hour");
	MAKE_INTEGER(value, st->tm_hour);
	ctkPutArray(arr, key, value);
	
	MAKE_STRING(key, "minute");
	MAKE_INTEGER(value, st->tm_min);
	ctkPutArray(arr, key, value);
	
	MAKE_STRING(key, "second");
	MAKE_INTEGER(value, st->tm_sec);
	ctkPutArray(arr, key, value);
	
	MAKE_STRING(key, "wday");
	MAKE_INTEGER(value, st->tm_wday);
	ctkPutArray(arr, key, value);
	
	MAKE_STRING(key, "yday");
	MAKE_INTEGER(value, st->tm_yday);
	ctkPutArray(arr, key, value);
	
	MAKE_STRING(key, "time");
	MAKE_INTEGER(value, t);
	ctkPutArray(arr, key, value);
	
	MAKE_ARRAY(value, arr);
	return value;
}
static CtkPrimitive p51("now,", currentTime);


static CtkObject currentTimeMillis(int nArgs, CtkObject*) {
    CtkObject result;
    MAKE_INTEGER(result, ctkGetTimeMillis());
    return result;
}
static CtkPrimitive p52("millisecond", currentTimeMillis);


static CtkObject chDir(int nArgs, CtkObject* args) {
	if (nArgs < 1 || !IS_STRING(args[0])) invalid_input();
    return chdir(args[0].u.svalue) == 0 ? ctkTrue : ctkFalse;
}
static CtkPrimitive p53("chdir", chDir);


static CtkObject pwd(int nArgs, CtkObject*) {
    char buf[1024];    
	    if (getcwd(buf, sizeof buf) == NULL) {
		return ctkNull;
    }
    CtkObject result;
    ALLOCATE_STRING(result, buf);
    return result;
}
static CtkPrimitive p54("pwd", pwd);


static CtkObject sleepSec(int nArgs, CtkObject* args) {
	if (nArgs < 1 || !IS_INTEGER(args[0])) invalid_input();
    CtkThread* thr = ctkGetCurrentThread();
    CtkMemoryManager::instance.endAccess(thr);
    ctkSleep(args[0].u.ivalue);
    CtkMemoryManager::instance.beginAccess(thr);
    return ctkNull;
}
static CtkPrimitive p55("sleep", sleepSec);


static CtkObject indexOf(int nArgs, CtkObject* args) {
    if (nArgs < 3 ||
		!IS_STRING(args[0]) ||
		!IS_STRING(args[1]) ||
		!IS_INTEGER(args[2])) invalid_input();
    char* s1 = args[0].u.svalue;
    char* s2 = args[1].u.svalue;
    int n1 = strlen(s1);
    int n2 = strlen(s2);
    CtkObject result;
    int i, j;
    for (i = 0; i <= n1-n2; i++) { 
		for (j = args[2].u.ivalue; j < n2 && s1[i+j] == s2[j]; j++);
		if (j == n2) { 
			MAKE_INTEGER(result, i);
			return result;
		}
    }
    MAKE_INTEGER(result, -1);
    return result;
}    
static CtkPrimitive p60("find", indexOf);


static CtkObject lastIndexOf(int nArgs, CtkObject* args) {
    if (nArgs < 2 ||
		!IS_STRING(args[0]) ||
		!IS_STRING(args[1])) invalid_input();
    char* s1 = args[0].u.svalue;
    char* s2 = args[1].u.svalue;
    int n1 = strlen(s1);
    int n2 = strlen(s2);
    CtkObject result;
    int i, j;
    for (i = n1-n2; i >= 0; i--) { 
		for (j = 0; j < n2 && s1[i+j] == s2[j]; j++);
		if (j == n2) { 
			MAKE_INTEGER(result, i);
			return result;
		}
    }
    MAKE_INTEGER(result, -1);
    return result;
}    
static CtkPrimitive p62("findlast", lastIndexOf);


static CtkObject substring(int nArgs, CtkObject* args) {
    if (nArgs < 2 ||
		!IS_STRING(args[0]) ||
		!IS_INTEGER(args[1])) invalid_input();
    char* src = args[0].u.svalue;
    unsigned len = strlen(src);
    unsigned from = args[1].u.ivalue;
    if (from > len) from = len;
    unsigned till = len;
    if (nArgs > 2) { 
		CHECK_PRIMITIVE_ARGUMENT(args[2], CTK_INTEGER);
		till = args[2].u.ivalue;
		if (till < from) { 
			till = from;
		} else if (till > len) { 
			till = len;
		}
    }
    char* dst = (char*)ctkAllocateObject(till-from+1, NULL);
    memcpy(dst, src + from, till-from);
    dst[till-from] = '\0';
    CtkObject result;
    MAKE_STRING(result, dst);
    return result;
}    
static CtkPrimitive p63("substring", substring);


static CtkObject toLower(int nArgs, CtkObject* args) {
    if (nArgs < 1 || !IS_STRING(args[0])) invalid_input();
    char* src = args[0].u.svalue;
    char* dst = (char*)ctkAllocateObject(strlen(src)+1, NULL);
    CtkObject result;
    MAKE_STRING(result, dst);
    while ((*dst++ = tolower((unsigned char)*src++)) != 0);
    return result;
}
static CtkPrimitive p64("tolower", toLower);


static CtkObject toUpper(int nArgs, CtkObject* args) {
    if (nArgs < 1 || !IS_STRING(args[0])) invalid_input();
    char* src = args[0].u.svalue;
    char* dst = (char*)ctkAllocateObject(strlen(src)+1, NULL);
    CtkObject result;
    MAKE_STRING(result, dst);
    while ((*dst++ = toupper((unsigned char)*src++)) != 0);
    return result;
}
static CtkPrimitive p65("toupper", toUpper);


static CtkObject trim(int nArgs, CtkObject* args) {
    if (nArgs < 1 || !IS_STRING(args[0])) invalid_input();
    char* src = args[0].u.svalue;
    int start = 0, end = strlen(src);
    while ((unsigned char)(src[start]-1) <= ' ') { 
        start += 1;
    }
    while (end > start && (unsigned char)src[end-1] <= ' ') { 
		end -= 1;
    }
    char* dst = (char*)ctkAllocateObject(end-start+1, NULL);
    CtkObject result;
    MAKE_STRING(result, dst);
    memcpy(dst, src+start, end-start);
    dst[end-start] = '\0';
    return result;
}
static CtkPrimitive p66("trim", trim);


static CtkObject startsWith(int nArgs, CtkObject* args) {
    if (nArgs < 2 || !IS_STRING(args[0]) || !IS_STRING(args[1])) invalid_input();
    char* p = args[0].u.svalue;
    char* q = args[1].u.svalue;
    while (*q != '\0') { 
		if (*p != *q) return ctkFalse;
    }
    return ctkTrue;
}
static CtkPrimitive p67("startswith", startsWith);


static CtkObject endsWith(int nArgs, CtkObject* args) {
    if (nArgs < 2 || !IS_STRING(args[0]) || !IS_STRING(args[1])) invalid_input();
    char* s1 = args[0].u.svalue;
    char* s2 = args[1].u.svalue;
    int n1 = strlen(s1);
    int n2 = strlen(s2);
    return (n1 >= n2 && strcmp(s1 + n1 - n2, s2) == 0) ? ctkTrue : ctkFalse;
}
static CtkPrimitive p68("endswith", endsWith);


static CtkObject string(int nArgs, CtkObject* args) {
    char buf[256];
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    if (args[0].type == CTK_STRING) return args[0];
    ctkToString(args[0], buf);
    char* str = (char*)ctkAllocateObject(strlen(buf) + 1, NULL);
    strcpy(str, buf);
    CtkObject result;
    MAKE_STRING(result, str);
    return result;
}
static CtkPrimitive p69("string", string);


static CtkObject onlySpaces(int nArgs, CtkObject* args) {
	if (nArgs < 1 || !IS_STRING(args[0])) invalid_input();
    char *s = args[0].u.svalue;
	for (int i = 0; s[i]; i++) {
		if (!isspace(s[i])) return ctkFalse;
	}
    return ctkTrue;
}
static CtkPrimitive p70("onlyspaces", onlySpaces);


static CtkObject onlyNumbers(int nArgs, CtkObject* args) {
	if (nArgs < 1 || !IS_STRING(args[0])) invalid_input();
	char *s = args[0].u.svalue;
	int i, len = strlen(s);
	for (i = 0; s[i]; i++) {
		if (!isspace(s[i])) break;
	}
	if (!s[i]) return ctkFalse;		// empty line
	i = 0;
	while (s[i]) {
		if (isdigit(s[i]) || isspace(s[i])) {
			i++;
			continue;
		}
		switch (s[i]) {
		case '-':case '+':
			if (isdigit(s[i+1]) || s[i+1] == '.' || s[i+1] == 'e' || s[i+1] == 'E') break;
			return ctkFalse;
		case 'e':case 'E':
			if (isdigit(s[i+1])) break;
			if (s[i+1] == '-' || s[i+1] == '+') {
				if (i+2 < len && isdigit(s[i+2])) break;
			}
			return ctkFalse;
		case '.':
			if (isdigit(s[i+1]) || s[i+1] == 'e' || s[i+1] == 'E') break;
			return ctkFalse;
		default:
			return ctkFalse;
		}
		i++;
	}
	return ctkTrue;
}
static CtkPrimitive p71("onlynumbers", onlyNumbers);


static CtkObject length(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    int len = 0;
    switch (args[0].type) { 
    case CTK_STRING:
		len = strlen(args[0].u.svalue);
		break;
    case CTK_ARRAY:
		len = ((CtkArray*)args[0].u.ptr)->nUsed;
		break;
    default:
		ctkThrowException("Incompatible type of primitive parameter");
    } 
    CtkObject result;
    MAKE_INTEGER(result, len);
    return result;
}
static CtkPrimitive p72("length", length);


static CtkObject integer(int nArgs, CtkObject* args) {
    CtkObject result;
    char ch;
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    switch (args[0].type) { 
    case CTK_INTEGER:
		return args[0];
	case CTK_REAL:
		MAKE_INTEGER(result, (ctk_integer)args[0].u.rvalue);
		break;
    case CTK_STRING:
		if (sscanf(args[0].u.svalue, "%ld%c", &result.u.ivalue, &ch) != 1) { 
			ctkThrowException("Invalid number format");
		}
		result.type = CTK_INTEGER;
		break;
    case CTK_RAW_POINTER:
		MAKE_INTEGER(result, (ctk_integer)args[0].u.ptr);
		break;        
    default: 
		ctkThrowException("Incompatible type of primitive parameter");
    }
    return result;
}
static CtkPrimitive p80("integer", integer);


static CtkObject real(int nArgs, CtkObject* args) {
    CtkObject result;
    char ch;
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    switch (args[0].type) { 
    case CTK_INTEGER:
		MAKE_REAL(result, args[0].u.ivalue);
		break;	
    case CTK_REAL:
		return args[0];
	case CTK_STRING:
		if (sscanf(args[0].u.svalue, "%lf%c", &result.u.rvalue, &ch) != 1) { 
			ctkThrowException("Invalid number format");
		}
		result.type = CTK_REAL;
		break;
    default: 
		ctkThrowException("Incompatible type of primitive parameter");
    }
    return result;
}
static CtkPrimitive p81("real", real);


static CtkObject isInteger(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    return args->type == CTK_INTEGER ? ctkTrue : ctkFalse;
}
static CtkPrimitive p81a("isInteger", isInteger);

static CtkObject isReal(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    return args->type == CTK_REAL ? ctkTrue : ctkFalse;
}
static CtkPrimitive p82("isreal", isReal);


static CtkObject isString(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    return args->type == CTK_STRING ? ctkTrue : ctkFalse;
}
static CtkPrimitive p83("isstring", isString);


static CtkObject isArray(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    return args->type == CTK_ARRAY ? ctkTrue : ctkFalse;
}
static CtkPrimitive p84("isarray", isArray);


static CtkObject isFile(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    return args->type == CTK_FILE ? ctkTrue : ctkFalse;
}
static CtkPrimitive p43("isfile", isFile);


static CtkObject isThread(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    return args->type == CTK_THREAD ? ctkTrue : ctkFalse;
}
static CtkPrimitive p85("isthread", isThread);


static CtkObject isNull(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    return args->type == CTK_NULL ? ctkTrue : ctkFalse;
}
static CtkPrimitive p86("isnull", isNull);


static CtkObject isMutex(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    return args->type == CTK_MUTEX ? ctkTrue : ctkFalse;
}
static CtkPrimitive p87("ismutex", isMutex);


static CtkObject isFunction(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    return args->type == CTK_FUNCTION ? ctkTrue : ctkFalse;
}
static CtkPrimitive p88("isfunction", isFunction);


static CtkObject isPrimitive(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    return args->type == CTK_PRIMITIVE ? ctkTrue : ctkFalse;
}
static CtkPrimitive p89("isprimitive", isPrimitive);


static CtkObject isPtr(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    return args->type == CTK_RAW_POINTER ? ctkTrue : ctkFalse;
}
static CtkPrimitive p90("isptr", isPtr);


static CtkObject isUsr(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    return args->type == CTK_USER_TYPE ? ctkTrue : ctkFalse;
}
static CtkPrimitive p91("isusr", isUsr);


static CtkObject getTypeName(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    static char const* typeMnem[] = { 
	"null", "integer", "real", "primitive", "raw_pointer", 
	"string", "array", "file", "mutex", "thread", "function", "frame", "user_type"
    };
    CtkObject result;
    ALLOCATE_STRING(result, typeMnem[args->type]);
    return result;
}
static CtkPrimitive p92("typename", getTypeName);


inline int cal2jul(int yy, int mm, int dd)
{
	yy -= mm < 3;
	mm += 12 * (mm < 3);
	int a = 365.25 * (yy + 4716),
		b = 30.6001 * (mm + 1);
	return a + b + dd - 2416556;
}

inline double jul2cal(int *yy, int *mm, int *dd, double jul)
{
	int a = jul + 2415032,
		b = a + 1524,
		c = (double(b) - 122.1) / 365.25,
		d = 365.25 * c,
		e = double(b - d) / 30.6001,
		f = 30.6001 * e;

	*dd = b - d - f;
	*mm = e - 1 - 12 * (e > 13);
	*yy = c - 4715 - (*mm > 2);

	return jul - floor(jul);
}

inline int jul2tim(int *hh, int *mn, int *sc, double jul)
{
	if (jul == 0)
	{
		*hh = 0; *mn = 0; *sc = 0;
		return 0;
	}
	jul = fabs(jul) * 24;
	*hh = jul;
	jul = (jul - *hh) * 60;
	*mn = jul;
	jul = (jul - *mn) * 60;
	*sc = jul;
	return (jul - *sc) * 1000;
}

static CtkObject CalToJul(int nArgs, CtkObject* args)
{
	if (nArgs < 3 ||
		!IS_NUMBER(args[0]) ||
		!IS_NUMBER(args[1]) ||
		!IS_NUMBER(args[2])) invalid_input();
	CtkObject ret;
	MAKE_INTEGER(ret, cal2jul(GET_NUMBER(args[0]), GET_NUMBER(args[1]), GET_NUMBER(args[2])));
	return ret;
}
static CtkPrimitive p100("caltojul", CalToJul);


static CtkObject JulToCal(int nArgs, CtkObject* args)
{
	if (nArgs < 1 || !IS_NUMBER(args[0])) invalid_input();
	
	int yy, mm, dd, hh, mn, sc, ms;
	double rm = jul2cal(&yy, &mm, &dd, GET_NUMBER(args[0]));
	ms = jul2tim(&hh, &mn, &sc, rm);

	CtkArray* arr = ctkCreateArray();
	CtkObject key, value;
	
	MAKE_STRING(key, "year");
	MAKE_INTEGER(value, yy);
	ctkPutArray(arr, key, value);

	MAKE_STRING(key, "month");
	MAKE_INTEGER(value, mm);
	ctkPutArray(arr, key, value);

	MAKE_STRING(key, "day");
	MAKE_INTEGER(value, dd);
	ctkPutArray(arr, key, value);

	MAKE_STRING(key, "hour");
	MAKE_INTEGER(value, hh);
	ctkPutArray(arr, key, value);

	MAKE_STRING(key, "minute");
	MAKE_INTEGER(value, mn);
	ctkPutArray(arr, key, value);

	MAKE_STRING(key, "second");
	MAKE_INTEGER(value, sc);
	ctkPutArray(arr, key, value);

	MAKE_ARRAY(value, arr);
	return value;
}
static CtkPrimitive p101("jultocal", JulToCal);


static CtkObject doGC(int nArgs, CtkObject* args)
{
	CtkMemoryManager::instance.doGC();
	return ctkTrue;
}
static CtkPrimitive p102("gc", doGC);


static CtkObject assertFunc(int nArgs, CtkObject* args) {
    CHECK_PRIMITIVE_N_ARGS(1, nArgs);
    if (!IS_TRUE(args[0])) ctkThrowException("Assertion failed");
    return ctkNull;
} 
static CtkPrimitive p103("assert", assertFunc);


static CtkObject printExceptionStackTrace(int nArgs, CtkObject*) {
    CHECK_PRIMITIVE_N_ARGS(0, nArgs);
    CtkThread* thread = ctkGetCurrentThread();
    if (thread->exceptionTrace != NULL) ctkPrintStackTrace(thread->exceptionTrace);
    return ctkNull;
}
static CtkPrimitive p104("printExceptionStackTrace", printExceptionStackTrace);


static CtkObject loadDLL(int nArgs, CtkObject* args) {
	if (nArgs < 1 || !IS_STRING(args[0])) invalid_input();
    return ctkLoadLibrary(args[0].u.svalue) ? ctkTrue : ctkFalse;
} 
static CtkPrimitive p105("loadlib", loadDLL);


static CtkObject loadModule(int nArgs, CtkObject* args) {
	if (nArgs < 1 || !IS_STRING(args[0])) invalid_input();
    return CtkCompiler::instance.compile(args[0].u.svalue) != NULL ? ctkTrue : ctkFalse;
} 
static CtkPrimitive p106("loadmodule", loadModule);


static CtkObject AddPath(int nArgs, CtkObject* args) {
    for (int i = 1; i <= nArgs; i++) {
		if (!IS_STRING(args[i-1])) invalid_input();
	    CtkCompiler::instance.path = new CtkPath(args[i-1].u.svalue, CtkCompiler::instance.path);
    }
    return ctkNull;
}
static CtkPrimitive p107("addpath", AddPath);


static CtkObject ShellExec(int nArgs, CtkObject* args) {
	if (nArgs < 3 ||
		!IS_INTEGER(args[0]) ||
		!IS_STRING(args[1]) ||
		!IS_STRING(args[2])) invalid_input();
	int flag =SW_SHOWNORMAL;
	if (args[0].u.ivalue < 1) flag =SW_HIDE;
	char *opr = args[1].u.svalue;
	char *cmd = args[2].u.svalue;;
	char *par = 0;
	char *dir = 0;
	if (nArgs > 3 && IS_STRING(args[3])) par = args[3].u.svalue;
	if (nArgs > 4 && IS_STRING(args[4])) dir = args[4].u.svalue;
	ShellExecute(0, opr, cmd, par, dir, flag);
    return ctkNull;
}
static CtkPrimitive p108("shell", ShellExec);


static CtkObject bms2utf(int nArgs, CtkObject* args) {
	if (nArgs < 1 || !IS_STRING(args[0])) invalid_input();
	char* s = args[0].u.svalue;
	unsigned char* byte = (unsigned char*)s;
	int len = strlen(s);
	char *utf8 = new char[4 * len + 1];
	unsigned short wc;
	char u[12];
	int i, j, n = 0, m;
	for (i = 0; i < len; i++) {
		if (s[i] > 0) {
			utf8[n] = s[i];
			n++;
		}
		else {
			wc = (byte[i] << 8) + byte[i+1];
			m = WideCharToMultiByte(CP_UTF8, 0, &wc, 1, u, 12, 0, 0);
			if (m <= 0) {
				delete[] utf8;
				ctkThrowException(" WideCharToMultiByte() error!");
			}
			for (j = 0; j < m; j++) {
				utf8[n] = u[j];
				n++;
			}
			i++;
		}
	}
	utf8[n] = 0;
	s = (char*)ctkAllocateObject(n+1, NULL);
	CHECK_ALLOC(s);
	strcpy(s, utf8);
	delete[] utf8;
	CtkObject ret;
	MAKE_STRING(ret, s);
	return ret;
}
static CtkPrimitive p109("bms2utf", bms2utf);


static CtkObject utf2mbs(int nArgs, CtkObject* args) {
	if (nArgs < 1 || !IS_STRING(args[0])) invalid_input();
	char* utf8 = args[0].u.svalue;
	int len = strlen(utf8);
	char *s = new char[len + 1];
	unsigned short wc;
	char *c;
	int i, nb = 1, n = 0, m;

	for (i = 0; i < len; i+= nb) {
		if      (utf8[i] > 0)                nb = 1;
		else if ((utf8[i] & 0x0F0) == 0x0E0) nb = 3;
		else if ((utf8[i] & 0x0E0) == 0x0C0) nb = 2;
		else                                 nb = 0;
		if (!nb) {
			delete[] s;
			ctkThrowException(" Invalid UTF-8 string!");
		}
		if (1 == nb) {
			s[n] = utf8[i];
			n++;
		}
		else {
			m = MultiByteToWideChar(CP_UTF8, 0, utf8+i, nb, &wc, 1);
			if (m <= 0) {
				delete[] s;
				ctkThrowException(" MultiByteToWideChar() error!");
			}
			c = (char*)&wc;
			s[n] = c[1];
			n++;
			s[n] = c[0];
			n++;
		}
	}
	s[n] = 0;
	c = (char*)ctkAllocateObject(n+1, NULL);
	CHECK_ALLOC(c);
	strcpy(c, s);
	delete[] s;
	CtkObject ret;
	MAKE_STRING(ret, c);
	return ret;
}
static CtkPrimitive p110("utf2mbs", utf2mbs);

