#include "amath.h"
#include "array.h"
#include "arraytmp.h"
#include <math.h>

#pragma warning(once: 4244)

#define DEG_TO_RAD		0.017453292519943 
#define RAD_TO_DEG		57.295779513082300 

int
amath_cal2jul(int yy, int mm, int dd)
{
	yy -= mm < 3;
	mm += 12 * (mm < 3);
	int a = 365.25 * (yy + 4716),
		b = 30.6001 * (mm + 1);
	return a + b + dd - 2416556;
}

double
amath_jul2cal(int *yy, int *mm, int *dd, double jul)
{
	int a = jul + 2415032,
		b = a + 1524,
		c = (double(b) - 122.1) / 365.25,
		d = 365.25 * c,
		e = double(b - d) / 30.6001,
		f = 30.6001 * e;

	*dd = b - d - f;
	*mm = e - 1 - 12 * (e > 13);
	*yy = c - 4715 - (*mm > 2);

	return jul - floor(jul);
}

int
amath_jul2tim(int *hh, int *mn, int *sc, double jul)
{
	if (jul == 0)
	{
		*hh = 0; *mn = 0; *sc = 0;
		return 0;
	}
	jul = fabs(jul) * 24;
	*hh = jul;
	jul = (jul - *hh) * 60;
	*mn = jul;
	jul = (jul - *mn) * 60;
	*sc = jul;
	return (jul - *sc) * 1000;
}

int
amath_make_julian(void *ptr, size_t n, int ds, int yy, int mm, int dd, int hh, int mn, int sc)
{
	if (!array_is_double(ptr)) return 0;
	zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
	if (arr->resize(n, 1) == 0) return 0;
	double jd0 = amath_cal2jul(yy, mm, dd);
	int sc0 = 3600 * hh + 60 * mn + sc;
	for (int i = 0; i < n; i++) (*arr)(i) = jd0 + double(sc0 + i * ds) / 86400;
	return 1;
}

int
amath_cos(void *ptr)
{
	if (array_is_double(ptr) == 0) return 0;
	zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
	size_t i, n = arr->ndat();
	for (i = 0; i < n; i++) (*arr)(i) = cos((*arr)(i));
	return 1;
}

int
amath_acos(void *ptr)
{
	if (array_is_double(ptr) == 0) return 0;
	zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
	size_t i, n = arr->ndat();
	for (i = 0; i < n; i++) (*arr)(i) = acos((*arr)(i));
	return 1;
}

int
amath_cosh(void *ptr)
{
	if (array_is_double(ptr) == 0) return 0;
	zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
	size_t i, n = arr->ndat();
	for (i = 0; i < n; i++) (*arr)(i) = cosh((*arr)(i));
	return 1;
}

int
amath_sin(void *ptr)
{
	if (array_is_double(ptr) == 0) return 0;
	zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
	size_t i, n = arr->ndat();
	for (i = 0; i < n; i++) (*arr)(i) = sin((*arr)(i));
	return 1;
}

int
amath_asin(void *ptr)
{
	if (array_is_double(ptr) == 0) return 0;
	zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
	size_t i, n = arr->ndat();
	for (i = 0; i < n; i++) (*arr)(i) = asin((*arr)(i));
	return 1;
}

int
amath_sinh(void *ptr)
{
	if (array_is_double(ptr) == 0) return 0;
	zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
	size_t i, n = arr->ndat();
	for (i = 0; i < n; i++) (*arr)(i) = sinh((*arr)(i));
	return 1;
}

int
amath_tan(void *ptr)
{
	if (array_is_double(ptr) == 0) return 0;
	zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
	size_t i, n = arr->ndat();
	for (i = 0; i < n; i++) (*arr)(i) = tan((*arr)(i));
	return 1;
}

int
amath_atan(void *ptr)
{
	if (array_is_double(ptr) == 0) return 0;
	zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
	size_t i, n = arr->ndat();
	for (i = 0; i < n; i++) (*arr)(i) = atan((*arr)(i));
	return 1;
}

int
amath_tanh(void *ptr)
{
	if (array_is_double(ptr) == 0) return 0;
	zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
	size_t i, n = arr->ndat();
	for (i = 0; i < n; i++) (*arr)(i) = tanh((*arr)(i));
	return 1;
}

int
amath_sqrt(void *ptr)
{
	if (array_is_double(ptr) == 0) return 0;
	zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
	size_t i, n = arr->ndat();
	for (i = 0; i < n; i++) (*arr)(i) = sqrt((*arr)(i));
	return 1;
}

int
amath_exp(void *ptr)
{
	if (array_is_double(ptr) == 0) return 0;
	zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
	size_t i, n = arr->ndat();
	for (i = 0; i < n; i++) (*arr)(i) = exp((*arr)(i));
	return 1;
}

int
amath_log(void *ptr)
{
	if (array_is_double(ptr) == 0) return 0;
	zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
	size_t i, n = arr->ndat();
	for (i = 0; i < n; i++)
	{
		if ((*arr)(i) <= 0)
		{
			array_set_error_message("amath_log() error.");
			return 0;
		}
		(*arr)(i) = log((*arr)(i));
	}
	return 1;
}

int
amath_log10(void *ptr)
{
	if (array_is_double(ptr) == 0) return 0;
	zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
	size_t i, n = arr->ndat();
	for (i = 0; i < n; i++)
	{
		if ((*arr)(i) <= 0)
		{
			array_set_error_message("amath_log10() error.");
			return 0;
		}
		(*arr)(i) = log10((*arr)(i));
	}
	return 1;
}

int
amath_ceil(void *ptr)
{
	if (array_is_double(ptr) == 0) return 0;
	zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
	size_t i, n = arr->ndat();
	for (i = 0; i < n; i++) (*arr)(i) = ceil((*arr)(i));
	return 1;
}

int
amath_floor(void *ptr)
{
	if (array_is_double(ptr) == 0) return 0;
	zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
	size_t i, n = arr->ndat();
	for (i = 0; i < n; i++) (*arr)(i) = floor((*arr)(i));
	return 1;
}

int
amath_abs(void *ptr)
{
	if (array_is_double(ptr) == 0) return 0;
	zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
	size_t i, n = arr->ndat();
	for (i = 0; i < n; i++) (*arr)(i) = fabs((*arr)(i));
	return 1;
}

int
amath_mod(void *ptr, double v)
{
	if (v == 0 || array_is_double(ptr) == 0) return 0;
	zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
	size_t i, n = arr->ndat();
	for (i = 0; i < n; i++) (*arr)(i) = fmod((*arr)(i), v);
	return 1;
}

int
amath_mod_ex(void *ptr, double v)
{
	if (v == 0 || array_is_double(ptr) == 0) return 0;
	zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
	size_t i, n = arr->ndat();
	for (i = 0; i < n; i++) {
		if ((*arr)(i) == 0) {
			array_set_error_message("Divided by zero.");
			return 0;
		}
		(*arr)(i) = fmod(v, (*arr)(i));
	}
	return 1;
}

int
amath_mod2(void *ptr, const void *ptr2)
{
	if (array_is_double(ptr) == 0) return 0;
	if (array_is_double(ptr2) == 0) return 0;
	zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
	const zeDouble *arr2 = reinterpret_cast<const zeDouble*>(ptr2);
	size_t i, n = arr->ndat();
	for (i = 0; i < n; i++) {
		if ((*arr2)(i) == 0) {
			array_set_error_message("Divided by zero.");
			return 0;
		}
		(*arr)(i) = fmod((*arr)(i), (*arr2)(i));
	}
	return 1;
}

int
amath_pow(void *ptr, double v)
{
	if (array_is_double(ptr) == 0) return 0;
	zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
	size_t i, n = arr->ndat();
	for (i = 0; i < n; i++) (*arr)(i) = pow((*arr)(i), v);
	return 1;
}

int
amath_pow_ex(void *ptr, double v)
{
	if (array_is_double(ptr) == 0) return 0;
	zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
	size_t i, n = arr->ndat();
	for (i = 0; i < n; i++) (*arr)(i) = pow(v, (*arr)(i));
	return 1;
}

int
amath_pow2(void *ptr, const void *ptr2)
{
	if (array_is_double(ptr) == 0) return 0;
	if (array_is_double(ptr2) == 0) return 0;
	zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
	const zeDouble *arr2 = reinterpret_cast<const zeDouble*>(ptr2);
	size_t i, n = arr->ndat();
	for (i = 0; i < n; i++) (*arr)(i) = pow((*arr)(i), (*arr2)(i));
	return 1;
}

int
amath_rand(void *ptr, long seed)
{
	if (array_is_double(ptr) == 0) return 0;
	zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
	size_t i, n = arr->ndat();
	seed = abs(seed);
	seed ^= 123459876;
	for (i = 0; i < n; i++) {
		long m = seed / 127773;
		seed = 16807 * (seed - 127773 * m) - 2836 * m;
		if (seed < 0) seed += 2147483647;
		(*arr)(i) = double(seed) / 2147483647;
	}
	return 1;
}

int
amath_make_identity(void *ptr, size_t n)
{
	if (array_is_double(ptr) == 0) return 0;
	zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
	arr->resize(n, n);
	size_t i, n2 = n * n;
	for (i = 0; i < n2; i++) (*arr)(i) = 0;
	for (i = 0; i < n; i++) (*arr) (i, i) = 1;
	return 1;
}

int
amath_make_rotatex(void *ptr, double deg)
{
	if (amath_make_identity(ptr, 4) == 0) return 0;
	zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
	double a = deg * DEG_TO_RAD, ct = cos(a), st = sin(a);
	(*arr)(1, 1) =  ct;
	(*arr)(1, 2) = -st;
	(*arr)(2, 1) =  st;
	(*arr)(2, 2) =  ct;
	return 1;
}

int
amath_make_rotatey(void *ptr, double deg)
{
	if (amath_make_identity(ptr, 4) == 0) return 0;
	zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
	double a = deg * DEG_TO_RAD, ct = cos(a), st = sin(a);
	(*arr)(0, 0) =  ct;
	(*arr)(0, 2) =  st;
	(*arr)(2, 0) = -st;
	(*arr)(2, 2) =  ct;
	return 1;
}

int
amath_make_rotatez(void *ptr, double deg)
{
	if (amath_make_identity(ptr, 4) == 0) return 0;
	zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
	double a = deg * DEG_TO_RAD, ct = cos(a), st = sin(a);
	(*arr)(0, 0) =  ct;
	(*arr)(0, 1) = -st;
	(*arr)(1, 0) =  st;
	(*arr)(1, 1) =  ct;
	return 1;
}

int
amath_make_scale(void *ptr, double sx, double sy, double sz)
{
	if (amath_make_identity(ptr, 4) == 0) return 0;
	zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
	(*arr)(0, 0) = sx;
	(*arr)(1, 1) = sy;
	(*arr)(2, 2) = sz;
	return 1;
}

int
amath_make_translate(void *ptr, double dx, double dy, double dz)
{
	if (amath_make_identity(ptr, 4) == 0) return 0;
	zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
	(*arr)(0, 3) = dx;
	(*arr)(1, 3) = dy;
	(*arr)(2, 3) = dz;
	return 1;
}