#include "arraytmp.h"
#include "array.h"

#include <sstream>

using namespace std;

#pragma warning(once: 4244)

char __arr_error__[256];

/////////////////////////////////////////////////////////////////////
void
array_set_error_message(const char* msg)
{
	strcpy(__arr_error__, msg);
}

void
array_set_error_message2(const char* msg1, const char* msg2)
{
	sprintf(__arr_error__, "%s: %s", msg1, msg2);
}

const char*
array_get_error_message()
{
	return __arr_error__;
}

int
array_check_type(const void *ptr)
{
	const zeChar* arr = (const zeChar*)ptr;
	int type = arr->type();
	if (type > ARRAY_NULL && type < ARRAY_TOP) return type;
	strcpy(__arr_error__, "Not an array type.");
	return 0;
}

int
array_check_type2(const void *ptr1, const void *ptr2)
{
	int type1 = array_check_type(ptr1);
	if (!type1) return 0;
	int type2 = array_check_type(ptr2);
	if (!type2) return 0;
	if (type1 == type2) return type1;
	strcpy(__arr_error__, "Not the same array type.");
	return 0;
}

int
array_is_int8(const void *ptr)
{
	int type = array_check_type(ptr);
	if (type == ARRAY_CHAR) return type;
	strcpy(__arr_error__, "Not a int8 array.");
	return 0;
}

int
array_is_uint8(const void *ptr)
{
	int type = array_check_type(ptr);
	if (type == ARRAY_UCHAR) return type;
	strcpy(__arr_error__, "Not a uint8 array.");
	return 0;
}

int
array_is_int16(const void *ptr)
{
	int type = array_check_type(ptr);
	if (type == ARRAY_SHORT) return type;
	strcpy(__arr_error__, "Not a int16 array.");
	return 0;
}

int
array_is_uint16(const void *ptr)
{
	int type = array_check_type(ptr);
	if (type == ARRAY_USHORT) return type;
	strcpy(__arr_error__, "Not a uint16 array.");
	return 0;
}

int
array_is_int32(const void *ptr)
{
	int type = array_check_type(ptr);
	if (type == ARRAY_INT) return type;
	strcpy(__arr_error__, "Not a int32 array.");
	return 0;
}

int
array_is_uint32(const void *ptr)
{
	int type = array_check_type(ptr);
	if (type == ARRAY_UINT) return type;
	strcpy(__arr_error__, "Not a uint32 array.");
	return 0;
}

int
array_is_float(const void *ptr)
{
	int type = array_check_type(ptr);
	if (type == ARRAY_FLOAT) return type;
	strcpy(__arr_error__, "Not a float array.");
	return 0;
}

int
array_is_double(const void *ptr)
{
	int type = array_check_type(ptr);
	if (type == ARRAY_DOUBLE) return type;
	strcpy(__arr_error__, "Not a double array.");
	return 0;
}

int
array_is_complex(const void *ptr)
{
	int type = array_check_type(ptr);
	if (type == ARRAY_COMPLEX) return type;
	strcpy(__arr_error__, "Not a complex array.");
	return 0;
}

int
array_is_triplex(const void *ptr)
{
	int type = array_check_type(ptr);
	if (type == ARRAY_TRIPLEX) return type;
	strcpy(__arr_error__, "Not a triplex array.");
	return 0;
}

int
array_is_quadruple(const void *ptr)
{
	int type = array_check_type(ptr);
	if (type == ARRAY_QUADRUPLE) return type;
	strcpy(__arr_error__, "Not a quadruple array.");
	return 0;
}


/////////////////////////////////////////////////////////////////////
void *
array_create(const char *type, size_t nrow, size_t ncol)
{
	array_set_error_message("array_create() error.");

	if (strcmp(type, "double") == 0)
	{
		zeDouble *ptr = new zeDouble;
		if (nrow == 0 || ncol == 0) return ptr;
		if (ptr->resize(nrow, ncol) > 0) return ptr;
		else return 0;
	}
	if (strcmp(type, "float") == 0)
	{
		zeFloat *ptr = new zeFloat;
		if (nrow == 0 || ncol == 0) return ptr;
		if (ptr->resize(nrow, ncol) > 0) return ptr;
		else return 0;
	}
	if (strcmp(type, "int") == 0)
	{
		zeInt *ptr = new zeInt;
		if (nrow == 0 || ncol == 0) return ptr;
		if (ptr->resize(nrow, ncol) > 0) return ptr;
		else return 0;
	}
	if (strcmp(type, "uint") == 0)
	{
		zeUint *ptr = new zeUint;
		if (nrow == 0 || ncol == 0) return ptr;
		if (ptr->resize(nrow, ncol) > 0) return ptr;
		else return 0;
	}
	if (strcmp(type, "int8") == 0)
	{
		zeChar *ptr = new zeChar;
		if (nrow == 0 || ncol == 0) return ptr;
		if (ptr->resize(nrow, ncol) > 0) return ptr;
		else return 0;
	}
	if (strcmp(type, "uint8") == 0)
	{
		zeUchar *ptr = new zeUchar;
		if (nrow == 0 || ncol == 0) return ptr;
		if (ptr->resize(nrow, ncol) > 0) return ptr;
		else return 0;
	}
	if (strcmp(type, "int16") == 0)
	{
		zeShort *ptr = new zeShort;
		if (nrow == 0 || ncol == 0) return ptr;
		if (ptr->resize(nrow, ncol) > 0) return ptr;
		else return 0;
	}
	if (strcmp(type, "uint16") == 0)
	{
		zeUshort *ptr = new zeUshort;
		if (nrow == 0 || ncol == 0) return ptr;
		if (ptr->resize(nrow, ncol) > 0) return ptr;
		else return 0;
	}
	if (strcmp(type, "complex") == 0)
	{
		zeComplex *ptr = new zeComplex;
		if (nrow == 0 || ncol == 0) return ptr;
		if (ptr->resize(nrow, ncol) > 0) return ptr;
		else return 0;
	}
	if (strcmp(type, "triplex") == 0)
	{
		zeTriplex *ptr = new zeTriplex;
		if (nrow == 0 || ncol == 0) return ptr;
		if (ptr->resize(nrow, ncol) > 0) return ptr;
		else return 0;
	}
	if (strcmp(type, "quadruple") == 0)
	{
		zeQuadruple *ptr = new zeQuadruple;
		if (nrow == 0 || ncol == 0) return ptr;
		if (ptr->resize(nrow, ncol) > 0) return ptr;
		else return 0;
	}
	return 0;
}

int
array_free(void *ptr)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	switch (type)
	{
	case ARRAY_CHAR:
		delete reinterpret_cast<zeChar*>(ptr);
		return 1;
	case ARRAY_UCHAR:
		delete reinterpret_cast<zeUchar*>(ptr);
		return 1;
	case ARRAY_SHORT:
		delete reinterpret_cast<zeShort*>(ptr);
		return 1;
	case ARRAY_USHORT:
		delete reinterpret_cast<zeUshort*>(ptr);
		return 1;
	case ARRAY_INT:
		delete reinterpret_cast<zeInt*>(ptr);
		return 1;
	case ARRAY_UINT:
		delete reinterpret_cast<zeUint*>(ptr);
		return 1;
	case ARRAY_FLOAT:
		delete reinterpret_cast<zeFloat*>(ptr);
		return 1;
	case ARRAY_DOUBLE:
		delete reinterpret_cast<zeDouble*>(ptr);
		return 1;
	case ARRAY_COMPLEX:
		delete reinterpret_cast<zeComplex*>(ptr);
		return 1;
	case ARRAY_TRIPLEX:
		delete reinterpret_cast<zeTriplex*>(ptr);
		return 1;
	case ARRAY_QUADRUPLE:
		delete reinterpret_cast<zeQuadruple*>(ptr);
		return 1;
	}
	return 0;
}

const char*
array_get_typename(const void *ptr)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	switch (type)
	{
	case ARRAY_CHAR:     return "int8";
	case ARRAY_UCHAR:    return "uint8";
	case ARRAY_SHORT:    return "int16";
	case ARRAY_USHORT:   return "uint16";
	case ARRAY_INT:      return "int";
	case ARRAY_UINT:     return "uint";
	case ARRAY_FLOAT:    return "float";
	case ARRAY_DOUBLE:   return "double";
	case ARRAY_COMPLEX:  return "complex";
	case ARRAY_TRIPLEX:  return "triplex";
	case ARRAY_QUADRUPLE:return "quadruple";
	}
	return 0;
}

void*
array_get_ptr(void *ptr, size_t *esize, size_t *n)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	switch (type)
	{
	case ARRAY_CHAR:
		{
			zeChar* arr = reinterpret_cast<zeChar*>(ptr);
			*esize = 1;
			*n = arr->ndat();
			return arr->data();
		}
	case ARRAY_UCHAR:
		{
			zeUchar* arr = reinterpret_cast<zeUchar*>(ptr);
			*esize = 1;
			*n = arr->ndat();
			return arr->data();
		}
	case ARRAY_SHORT:
		{
			zeShort* arr = reinterpret_cast<zeShort*>(ptr);
			*esize = sizeof(short);
			*n = arr->ndat();
			return arr->data();
		}
	case ARRAY_USHORT:
		{
			zeUshort* arr = reinterpret_cast<zeUshort*>(ptr);
			*esize = sizeof(short);
			*n = arr->ndat();
			return arr->data();
		}
	case ARRAY_INT:
		{
			zeInt* arr = reinterpret_cast<zeInt*>(ptr);
			*esize = sizeof(int);
			*n = arr->ndat();
			return arr->data();
		}
	case ARRAY_UINT:
		{
			zeUint* arr = reinterpret_cast<zeUint*>(ptr);
			*esize = sizeof(int);
			*n = arr->ndat();
			return arr->data();
		}
	case ARRAY_FLOAT:
		{
			zeFloat* arr = reinterpret_cast<zeFloat*>(ptr);
			*esize = sizeof(float);
			*n = arr->ndat();
			return arr->data();
		}
	case ARRAY_DOUBLE:
		{
			zeDouble* arr = reinterpret_cast<zeDouble*>(ptr);
			*esize = sizeof(double);
			*n = arr->ndat();
			return arr->data();
		}
	case ARRAY_COMPLEX:
		{
			zeComplex* arr = reinterpret_cast<zeComplex*>(ptr);
			*esize = sizeof(sComplex);
			*n = arr->ndat();
			return arr->data();
		}
	case ARRAY_TRIPLEX:
		{
			zeTriplex* arr = reinterpret_cast<zeTriplex*>(ptr);
			*esize = sizeof(sTriplex);
			*n = arr->ndat();
			return arr->data();
		}
	case ARRAY_QUADRUPLE:
		{
			zeQuadruple* arr = reinterpret_cast<zeQuadruple*>(ptr);
			*esize = sizeof(sQuadruple);
			*n = arr->ndat();
			return arr->data();
		}
	}
	return 0;
}

int
array_get_size(const void *ptr, size_t *nrow, size_t *ncol)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	switch (type)
	{
	case ARRAY_CHAR:
		{
			const zeChar* arr = reinterpret_cast<const zeChar*>(ptr);
			*nrow = arr->nrow();
			*ncol = arr->ncol();
			return 1;
		}
	case ARRAY_UCHAR:
		{
			const zeUchar* arr = reinterpret_cast<const zeUchar*>(ptr);
			*nrow = arr->nrow();
			*ncol = arr->ncol();
			return 1;
		}
	case ARRAY_SHORT:
		{
			const zeShort* arr = reinterpret_cast<const zeShort*>(ptr);
			*nrow = arr->nrow();
			*ncol = arr->ncol();
			return 1;
		}
	case ARRAY_USHORT:
		{
			const zeUshort* arr = reinterpret_cast<const zeUshort*>(ptr);
			*nrow = arr->nrow();
			*ncol = arr->ncol();
			return 1;
		}
	case ARRAY_INT:
		{
			const zeInt* arr = reinterpret_cast<const zeInt*>(ptr);
			*nrow = arr->nrow();
			*ncol = arr->ncol();
			return 1;
		}
	case ARRAY_UINT:
		{
			const zeUint* arr = reinterpret_cast<const zeUint*>(ptr);
			*nrow = arr->nrow();
			*ncol = arr->ncol();
			return 1;
		}
	case ARRAY_FLOAT:
		{
			const zeFloat* arr = reinterpret_cast<const zeFloat*>(ptr);
			*nrow = arr->nrow();
			*ncol = arr->ncol();
			return 1;
		}
	case ARRAY_DOUBLE:
		{
			const zeDouble* arr = reinterpret_cast<const zeDouble*>(ptr);
			*nrow = arr->nrow();
			*ncol = arr->ncol();
			return 1;
		}
	case ARRAY_COMPLEX:
		{
			const zeComplex* arr = reinterpret_cast<const zeComplex*>(ptr);
			*nrow = arr->nrow();
			*ncol = arr->ncol();
			return 1;
		}
	case ARRAY_TRIPLEX:
		{
			const zeTriplex* arr = reinterpret_cast<const zeTriplex*>(ptr);
			*nrow = arr->nrow();
			*ncol = arr->ncol();
			return 1;
		}
	case ARRAY_QUADRUPLE:
		{
			const zeQuadruple* arr = reinterpret_cast<const zeQuadruple*>(ptr);
			*nrow = arr->nrow();
			*ncol = arr->ncol();
			return 1;
		}
	}
	return 0;
}

int
array_resize(void *ptr, size_t nrow, size_t ncol, bool keep)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	array_set_error_message("array_resize() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(ptr)->resize(nrow, ncol, keep);
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(ptr)->resize(nrow, ncol, keep);
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(ptr)->resize(nrow, ncol, keep);
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(ptr)->resize(nrow, ncol, keep);
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(ptr)->resize(nrow, ncol, keep);
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(ptr)->resize(nrow, ncol, keep);
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(ptr)->resize(nrow, ncol, keep);
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(ptr)->resize(nrow, ncol, keep);
	case ARRAY_COMPLEX:
		return reinterpret_cast<zeComplex*>(ptr)->resize(nrow, ncol, keep);
	case ARRAY_TRIPLEX:
		return reinterpret_cast<zeTriplex*>(ptr)->resize(nrow, ncol, keep);
	case ARRAY_QUADRUPLE:
		return reinterpret_cast<zeQuadruple*>(ptr)->resize(nrow, ncol, keep);
	}
	return 0;
}

int
array_reshape(void *ptr, size_t nrow, size_t ncol)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	array_set_error_message("array_reshape() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(ptr)->reshape(nrow, ncol);
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(ptr)->reshape(nrow, ncol);
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(ptr)->reshape(nrow, ncol);
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(ptr)->reshape(nrow, ncol);
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(ptr)->reshape(nrow, ncol);
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(ptr)->reshape(nrow, ncol);
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(ptr)->reshape(nrow, ncol);
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(ptr)->reshape(nrow, ncol);
	case ARRAY_COMPLEX:
		return reinterpret_cast<zeComplex*>(ptr)->reshape(nrow, ncol);
	case ARRAY_TRIPLEX:
		return reinterpret_cast<zeTriplex*>(ptr)->reshape(nrow, ncol);
	case ARRAY_QUADRUPLE:
		return reinterpret_cast<zeQuadruple*>(ptr)->reshape(nrow, ncol);
	}
	return 0;
}

int
array_fill(void *ptr, double v, double dv)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	array_set_error_message("array_range() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(ptr)->fill(v, dv);
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(ptr)->fill(v, dv);
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(ptr)->fill(v, dv);
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(ptr)->fill(v, dv);
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(ptr)->fill(v, dv);
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(ptr)->fill(v, dv);
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(ptr)->fill(v, dv);
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(ptr)->fill(v, dv);
	}
	return 0;
}

int
array_negate(void *ptr)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	array_set_error_message("array_negate() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(ptr)->negate();
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(ptr)->negate();
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(ptr)->negate();
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(ptr)->negate();
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(ptr)->negate();
	}
	return 0;
}

int
array_add(void *ptr, double v)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	if (v == 0) return 1;
	array_set_error_message("array_add() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(ptr)->add(v);
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(ptr)->add(v);
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(ptr)->add(v);
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(ptr)->add(v);
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(ptr)->add(v);
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(ptr)->add(v);
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(ptr)->add(v);
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(ptr)->add(v);
	}
	return 0;
}

int
array_add2(void *dst, const void *src)
{
	int type = array_check_type2(dst, src);
	if (!type) return 0;
	array_set_error_message("array_add2() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(dst)->add(*reinterpret_cast<const zeChar*>(src));
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(dst)->add(*reinterpret_cast<const zeUchar*>(src));
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(dst)->add(*reinterpret_cast<const zeShort*>(src));
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(dst)->add(*reinterpret_cast<const zeUshort*>(src));
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(dst)->add(*reinterpret_cast<const zeInt*>(src));
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(dst)->add(*reinterpret_cast<const zeUint*>(src));
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(dst)->add(*reinterpret_cast<const zeFloat*>(src));
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(dst)->add(*reinterpret_cast<const zeDouble*>(src));
	}
	return 0;
}

int
array_subtract(void *ptr, double v)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	if (v == 0) return 1;
	array_set_error_message("array_subtract() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(ptr)->subtract(v);
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(ptr)->subtract(v);
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(ptr)->subtract(v);
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(ptr)->subtract(v);
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(ptr)->subtract(v);
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(ptr)->subtract(v);
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(ptr)->subtract(v);
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(ptr)->subtract(v);
	}
	return 0;
}

int
array_subtract2(void *dst, const void *src)
{
	int type = array_check_type2(dst, src);
	if (!type) return 0;
	array_set_error_message("array_subtract2() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(dst)->subtract(*reinterpret_cast<const zeChar*>(src));
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(dst)->subtract(*reinterpret_cast<const zeUchar*>(src));
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(dst)->subtract(*reinterpret_cast<const zeShort*>(src));
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(dst)->subtract(*reinterpret_cast<const zeUshort*>(src));
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(dst)->subtract(*reinterpret_cast<const zeInt*>(src));
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(dst)->subtract(*reinterpret_cast<const zeUint*>(src));
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(dst)->subtract(*reinterpret_cast<const zeFloat*>(src));
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(dst)->subtract(*reinterpret_cast<const zeDouble*>(src));
	}
	return 0;
}

int
array_multiply(void *ptr, double v)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	if (v == 1) return 1;
	array_set_error_message("array_multiply() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(ptr)->multiply(v);
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(ptr)->multiply(v);
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(ptr)->multiply(v);
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(ptr)->multiply(v);
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(ptr)->multiply(v);
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(ptr)->multiply(v);
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(ptr)->multiply(v);
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(ptr)->multiply(v);
	}
	return 0;
}

int
array_multiply2(void *dst, const void *src)
{
	int type = array_check_type2(dst, src);
	if (!type) return 0;
	array_set_error_message("array_multiply2() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(dst)->multiply(*reinterpret_cast<const zeChar*>(src));
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(dst)->multiply(*reinterpret_cast<const zeUchar*>(src));
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(dst)->multiply(*reinterpret_cast<const zeShort*>(src));
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(dst)->multiply(*reinterpret_cast<const zeUshort*>(src));
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(dst)->multiply(*reinterpret_cast<const zeInt*>(src));
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(dst)->multiply(*reinterpret_cast<const zeUint*>(src));
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(dst)->multiply(*reinterpret_cast<const zeFloat*>(src));
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(dst)->multiply(*reinterpret_cast<const zeDouble*>(src));
	}
	return 0;
}

int
array_element_inverse(void *ptr)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	array_set_error_message("array_element_inverse() error.");
	switch (type)
	{
	case ARRAY_FLOAT:
		{
		zeFloat* arr = reinterpret_cast<zeFloat*>(ptr);
		int i, n = arr->ndat();
		for (i = 0; i < n; i++)
		{
			if ((*arr)(i) == 0) return 0;
			(*arr)(i) = 1.0 / (*arr)(i);
		}
		return 1;
		}
	case ARRAY_DOUBLE:
		{
		zeDouble* arr = reinterpret_cast<zeDouble*>(ptr);
		int i, n = arr->ndat();
		for (i = 0; i < n; i++)
		{
			if ((*arr)(i) == 0) return 0;
			(*arr)(i) = 1.0 / (*arr)(i);
		}
		return 1;
		}
	}
	return 0;
}


int
array_divide(void *ptr, double v)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	if (v == 1) return 1;
	array_set_error_message("array_divide() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(ptr)->divide(v);
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(ptr)->divide(v);
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(ptr)->divide(v);
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(ptr)->divide(v);
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(ptr)->divide(v);
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(ptr)->divide(v);
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(ptr)->divide(v);
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(ptr)->divide(v);
	}
	return 0;
}

int
array_divide2(void *dst, const void *src)
{
	int type = array_check_type2(dst, src);
	if (!type) return 0;
	array_set_error_message("array_divide2() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(dst)->divide(*reinterpret_cast<const zeChar*>(src));
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(dst)->divide(*reinterpret_cast<const zeUchar*>(src));
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(dst)->divide(*reinterpret_cast<const zeShort*>(src));
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(dst)->divide(*reinterpret_cast<const zeUshort*>(src));
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(dst)->divide(*reinterpret_cast<const zeInt*>(src));
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(dst)->divide(*reinterpret_cast<const zeUint*>(src));
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(dst)->divide(*reinterpret_cast<const zeFloat*>(src));
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(dst)->divide(*reinterpret_cast<const zeDouble*>(src));
	}
	return 0;
}

int
array_and(void *idx, const void *ptr1, const void *ptr2)
{
	int type = array_is_int8(idx);
	if (!type) return 0;
	type = array_check_type2(ptr1, ptr2);
	if (!type) return 0;
	zeChar* dst = reinterpret_cast<zeChar*>(idx);
	array_set_error_message("array_and() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<const zeChar*>(ptr1)->and(*dst, *reinterpret_cast<const zeChar*>(ptr2));
	case ARRAY_UCHAR:
		return reinterpret_cast<const zeUchar*>(ptr1)->and(*dst, *reinterpret_cast<const zeUchar*>(ptr2));
	case ARRAY_SHORT:
		return reinterpret_cast<const zeShort*>(ptr1)->and(*dst, *reinterpret_cast<const zeShort*>(ptr2));
	case ARRAY_USHORT:
		return reinterpret_cast<const zeUshort*>(ptr1)->and(*dst, *reinterpret_cast<const zeUshort*>(ptr2));
	case ARRAY_INT:
		return reinterpret_cast<const zeInt*>(ptr1)->and(*dst, *reinterpret_cast<const zeInt*>(ptr2));
	case ARRAY_UINT:
		return reinterpret_cast<const zeUint*>(ptr1)->and(*dst, *reinterpret_cast<const zeUint*>(ptr2));
	case ARRAY_FLOAT:
		return reinterpret_cast<const zeFloat*>(ptr1)->and(*dst, *reinterpret_cast<const zeFloat*>(ptr2));
	case ARRAY_DOUBLE:
		return reinterpret_cast<const zeDouble*>(ptr1)->and(*dst, *reinterpret_cast<const zeDouble*>(ptr2));
	}
	return 0;
}

int
array_or(void *idx, const void *ptr1, const void *ptr2)
{
	int type = array_is_int8(idx);
	if (!type) return 0;
	type = array_check_type2(ptr1, ptr2);
	if (!type) return 0;
	zeChar* dst = reinterpret_cast<zeChar*>(idx);
	array_set_error_message("array_or() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<const zeChar*>(ptr1)->or(*dst, *reinterpret_cast<const zeChar*>(ptr2));
	case ARRAY_UCHAR:
		return reinterpret_cast<const zeUchar*>(ptr1)->or(*dst, *reinterpret_cast<const zeUchar*>(ptr2));
	case ARRAY_SHORT:
		return reinterpret_cast<const zeShort*>(ptr1)->or(*dst, *reinterpret_cast<const zeShort*>(ptr2));
	case ARRAY_USHORT:
		return reinterpret_cast<const zeUshort*>(ptr1)->or(*dst, *reinterpret_cast<const zeUshort*>(ptr2));
	case ARRAY_INT:
		return reinterpret_cast<const zeInt*>(ptr1)->or(*dst, *reinterpret_cast<const zeInt*>(ptr2));
	case ARRAY_UINT:
		return reinterpret_cast<const zeUint*>(ptr1)->or(*dst, *reinterpret_cast<const zeUint*>(ptr2));
	case ARRAY_FLOAT:
		return reinterpret_cast<const zeFloat*>(ptr1)->or(*dst, *reinterpret_cast<const zeFloat*>(ptr2));
	case ARRAY_DOUBLE:
		return reinterpret_cast<const zeDouble*>(ptr1)->or(*dst, *reinterpret_cast<const zeDouble*>(ptr2));
	}
	return 0;
}

int
array_xor(void *idx, const void *ptr1, const void *ptr2)
{
	int type = array_is_int8(idx);
	if (!type) return 0;
	type = array_check_type2(ptr1, ptr2);
	if (!type) return 0;
	zeChar* dst = reinterpret_cast<zeChar*>(idx);
	array_set_error_message("array_xor() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<const zeChar*>(ptr1)->xor(*dst, *reinterpret_cast<const zeChar*>(ptr2));
	case ARRAY_UCHAR:
		return reinterpret_cast<const zeUchar*>(ptr1)->xor(*dst, *reinterpret_cast<const zeUchar*>(ptr2));
	case ARRAY_SHORT:
		return reinterpret_cast<const zeShort*>(ptr1)->xor(*dst, *reinterpret_cast<const zeShort*>(ptr2));
	case ARRAY_USHORT:
		return reinterpret_cast<const zeUshort*>(ptr1)->xor(*dst, *reinterpret_cast<const zeUshort*>(ptr2));
	case ARRAY_INT:
		return reinterpret_cast<const zeInt*>(ptr1)->xor(*dst, *reinterpret_cast<const zeInt*>(ptr2));
	case ARRAY_UINT:
		return reinterpret_cast<const zeUint*>(ptr1)->xor(*dst, *reinterpret_cast<const zeUint*>(ptr2));
	case ARRAY_FLOAT:
		return reinterpret_cast<const zeFloat*>(ptr1)->xor(*dst, *reinterpret_cast<const zeFloat*>(ptr2));
	case ARRAY_DOUBLE:
		return reinterpret_cast<const zeDouble*>(ptr1)->xor(*dst, *reinterpret_cast<const zeDouble*>(ptr2));
	}
	return 0;
}

int
array_equal(void *idx, const void *ptr, double v)
{
	int type = array_is_int8(idx);
	if (!type) return 0;
	type = array_check_type(ptr);
	if (!type) return 0;
	zeChar* dst = reinterpret_cast<zeChar*>(idx);
	array_set_error_message("array_equal() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<const zeChar*>(ptr)->equal(*dst, v);
	case ARRAY_UCHAR:
		return reinterpret_cast<const zeUchar*>(ptr)->equal(*dst, v);
	case ARRAY_SHORT:
		return reinterpret_cast<const zeShort*>(ptr)->equal(*dst, v);
	case ARRAY_USHORT:
		return reinterpret_cast<const zeUshort*>(ptr)->equal(*dst, v);
	case ARRAY_INT:
		return reinterpret_cast<const zeInt*>(ptr)->equal(*dst, v);
	case ARRAY_UINT:
		return reinterpret_cast<const zeUint*>(ptr)->equal(*dst, v);
	case ARRAY_FLOAT:
		return reinterpret_cast<const zeFloat*>(ptr)->equal(*dst, v);
	case ARRAY_DOUBLE:
		return reinterpret_cast<const zeDouble*>(ptr)->equal(*dst, v);
	}
	return 0;
}

int
array_equal2(void *idx, const void *ptr1, const void *ptr2)
{
	int type = array_is_int8(idx);
	if (!type) return 0;
	type = array_check_type2(ptr1, ptr2);
	if (!type) return 0;
	zeChar* dst = reinterpret_cast<zeChar*>(idx);
	array_set_error_message("array_equal2() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<const zeChar*>(ptr1)->equal(*dst, *reinterpret_cast<const zeChar*>(ptr2));
	case ARRAY_UCHAR:
		return reinterpret_cast<const zeUchar*>(ptr1)->equal(*dst, *reinterpret_cast<const zeUchar*>(ptr2));
	case ARRAY_SHORT:
		return reinterpret_cast<const zeShort*>(ptr1)->equal(*dst, *reinterpret_cast<const zeShort*>(ptr2));
	case ARRAY_USHORT:
		return reinterpret_cast<const zeUshort*>(ptr1)->equal(*dst, *reinterpret_cast<const zeUshort*>(ptr2));
	case ARRAY_INT:
		return reinterpret_cast<const zeInt*>(ptr1)->equal(*dst, *reinterpret_cast<const zeInt*>(ptr2));
	case ARRAY_UINT:
		return reinterpret_cast<const zeUint*>(ptr1)->equal(*dst, *reinterpret_cast<const zeUint*>(ptr2));
	case ARRAY_FLOAT:
		return reinterpret_cast<const zeFloat*>(ptr1)->equal(*dst, *reinterpret_cast<const zeFloat*>(ptr2));
	case ARRAY_DOUBLE:
		return reinterpret_cast<const zeDouble*>(ptr1)->equal(*dst, *reinterpret_cast<const zeDouble*>(ptr2));
	}
	return 0;
}

int
array_nequal(void *idx, const void *ptr, double v)
{
	int type = array_is_int8(idx);
	if (!type) return 0;
	type = array_check_type(ptr);
	if (!type) return 0;
	zeChar* dst = reinterpret_cast<zeChar*>(idx);
	array_set_error_message("array_nequal() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<const zeChar*>(ptr)->nequal(*dst, v);
	case ARRAY_UCHAR:
		return reinterpret_cast<const zeUchar*>(ptr)->nequal(*dst, v);
	case ARRAY_SHORT:
		return reinterpret_cast<const zeShort*>(ptr)->nequal(*dst, v);
	case ARRAY_USHORT:
		return reinterpret_cast<const zeUshort*>(ptr)->nequal(*dst, v);
	case ARRAY_INT:
		return reinterpret_cast<const zeInt*>(ptr)->nequal(*dst, v);
	case ARRAY_UINT:
		return reinterpret_cast<const zeUint*>(ptr)->nequal(*dst, v);
	case ARRAY_FLOAT:
		return reinterpret_cast<const zeFloat*>(ptr)->nequal(*dst, v);
	case ARRAY_DOUBLE:
		return reinterpret_cast<const zeDouble*>(ptr)->nequal(*dst, v);
	}
	return 0;
}

int
array_nequal2(void *idx, const void *ptr1, const void *ptr2)
{
	int type = array_is_int8(idx);
	if (!type) return 0;
	type = array_check_type2(ptr1, ptr2);
	if (!type) return 0;
	zeChar* dst = reinterpret_cast<zeChar*>(idx);
	array_set_error_message("array_nequal2() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<const zeChar*>(ptr1)->nequal(*dst, *reinterpret_cast<const zeChar*>(ptr2));
	case ARRAY_UCHAR:
		return reinterpret_cast<const zeUchar*>(ptr1)->nequal(*dst, *reinterpret_cast<const zeUchar*>(ptr2));
	case ARRAY_SHORT:
		return reinterpret_cast<const zeShort*>(ptr1)->nequal(*dst, *reinterpret_cast<const zeShort*>(ptr2));
	case ARRAY_USHORT:
		return reinterpret_cast<const zeUshort*>(ptr1)->nequal(*dst, *reinterpret_cast<const zeUshort*>(ptr2));
	case ARRAY_INT:
		return reinterpret_cast<const zeInt*>(ptr1)->nequal(*dst, *reinterpret_cast<const zeInt*>(ptr2));
	case ARRAY_UINT:
		return reinterpret_cast<const zeUint*>(ptr1)->nequal(*dst, *reinterpret_cast<const zeUint*>(ptr2));
	case ARRAY_FLOAT:
		return reinterpret_cast<const zeFloat*>(ptr1)->nequal(*dst, *reinterpret_cast<const zeFloat*>(ptr2));
	case ARRAY_DOUBLE:
		return reinterpret_cast<const zeDouble*>(ptr1)->nequal(*dst, *reinterpret_cast<const zeDouble*>(ptr2));
	}
	return 0;
}

int
array_less(void *idx, const void *ptr, double v)
{
	int type = array_is_int8(idx);
	if (!type) return 0;
	type = array_check_type(ptr);
	if (!type) return 0;
	zeChar* dst = reinterpret_cast<zeChar*>(idx);
	array_set_error_message("array_less() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<const zeChar*>(ptr)->less(*dst, v);
	case ARRAY_UCHAR:
		return reinterpret_cast<const zeUchar*>(ptr)->less(*dst, v);
	case ARRAY_SHORT:
		return reinterpret_cast<const zeShort*>(ptr)->less(*dst, v);
	case ARRAY_USHORT:
		return reinterpret_cast<const zeUshort*>(ptr)->less(*dst, v);
	case ARRAY_INT:
		return reinterpret_cast<const zeInt*>(ptr)->less(*dst, v);
	case ARRAY_UINT:
		return reinterpret_cast<const zeUint*>(ptr)->less(*dst, v);
	case ARRAY_FLOAT:
		return reinterpret_cast<const zeFloat*>(ptr)->less(*dst, v);
	case ARRAY_DOUBLE:
		return reinterpret_cast<const zeDouble*>(ptr)->less(*dst, v);
	}
	return 0;
}

int
array_less2(void *idx, const void *ptr1, const void *ptr2)
{
	int type = array_is_int8(idx);
	if (!type) return 0;
	type = array_check_type2(ptr1, ptr2);
	if (!type) return 0;
	zeChar* dst = reinterpret_cast<zeChar*>(idx);
	array_set_error_message("array_less2() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<const zeChar*>(ptr1)->less(*dst, *reinterpret_cast<const zeChar*>(ptr2));
	case ARRAY_UCHAR:
		return reinterpret_cast<const zeUchar*>(ptr1)->less(*dst, *reinterpret_cast<const zeUchar*>(ptr2));
	case ARRAY_SHORT:
		return reinterpret_cast<const zeShort*>(ptr1)->less(*dst, *reinterpret_cast<const zeShort*>(ptr2));
	case ARRAY_USHORT:
		return reinterpret_cast<const zeUshort*>(ptr1)->less(*dst, *reinterpret_cast<const zeUshort*>(ptr2));
	case ARRAY_INT:
		return reinterpret_cast<const zeInt*>(ptr1)->less(*dst, *reinterpret_cast<const zeInt*>(ptr2));
	case ARRAY_UINT:
		return reinterpret_cast<const zeUint*>(ptr1)->less(*dst, *reinterpret_cast<const zeUint*>(ptr2));
	case ARRAY_FLOAT:
		return reinterpret_cast<const zeFloat*>(ptr1)->less(*dst, *reinterpret_cast<const zeFloat*>(ptr2));
	case ARRAY_DOUBLE:
		return reinterpret_cast<const zeDouble*>(ptr1)->less(*dst, *reinterpret_cast<const zeDouble*>(ptr2));
	}
	return 0;
}

int
array_lequal(void *idx, const void *ptr, double v)
{
	int type = array_is_int8(idx);
	if (!type) return 0;
	type = array_check_type(ptr);
	if (!type) return 0;
	zeChar* dst = reinterpret_cast<zeChar*>(idx);
	array_set_error_message("array_lequal() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<const zeChar*>(ptr)->lequal(*dst, v);
	case ARRAY_UCHAR:
		return reinterpret_cast<const zeUchar*>(ptr)->lequal(*dst, v);
	case ARRAY_SHORT:
		return reinterpret_cast<const zeShort*>(ptr)->lequal(*dst, v);
	case ARRAY_USHORT:
		return reinterpret_cast<const zeUshort*>(ptr)->lequal(*dst, v);
	case ARRAY_INT:
		return reinterpret_cast<const zeInt*>(ptr)->lequal(*dst, v);
	case ARRAY_UINT:
		return reinterpret_cast<const zeUint*>(ptr)->lequal(*dst, v);
	case ARRAY_FLOAT:
		return reinterpret_cast<const zeFloat*>(ptr)->lequal(*dst, v);
	case ARRAY_DOUBLE:
		return reinterpret_cast<const zeDouble*>(ptr)->lequal(*dst, v);
	}
	return 0;
}

int
array_lequal2(void *idx, const void *ptr1, const void *ptr2)
{
	int type = array_is_int8(idx);
	if (!type) return 0;
	type = array_check_type2(ptr1, ptr2);
	if (!type) return 0;
	zeChar* dst = reinterpret_cast<zeChar*>(idx);
	array_set_error_message("array_lequal2() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<const zeChar*>(ptr1)->lequal(*dst, *reinterpret_cast<const zeChar*>(ptr2));
	case ARRAY_UCHAR:
		return reinterpret_cast<const zeUchar*>(ptr1)->lequal(*dst, *reinterpret_cast<const zeUchar*>(ptr2));
	case ARRAY_SHORT:
		return reinterpret_cast<const zeShort*>(ptr1)->lequal(*dst, *reinterpret_cast<const zeShort*>(ptr2));
	case ARRAY_USHORT:
		return reinterpret_cast<const zeUshort*>(ptr1)->lequal(*dst, *reinterpret_cast<const zeUshort*>(ptr2));
	case ARRAY_INT:
		return reinterpret_cast<const zeInt*>(ptr1)->lequal(*dst, *reinterpret_cast<const zeInt*>(ptr2));
	case ARRAY_UINT:
		return reinterpret_cast<const zeUint*>(ptr1)->lequal(*dst, *reinterpret_cast<const zeUint*>(ptr2));
	case ARRAY_FLOAT:
		return reinterpret_cast<const zeFloat*>(ptr1)->lequal(*dst, *reinterpret_cast<const zeFloat*>(ptr2));
	case ARRAY_DOUBLE:
		return reinterpret_cast<const zeDouble*>(ptr1)->lequal(*dst, *reinterpret_cast<const zeDouble*>(ptr2));
	}
	return 0;
}

int
array_greater(void *idx, const void *ptr, double v)
{
	int type = array_is_int8(idx);
	if (!type) return 0;
	type = array_check_type(ptr);
	if (!type) return 0;
	zeChar* dst = reinterpret_cast<zeChar*>(idx);
	array_set_error_message("array_greater() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<const zeChar*>(ptr)->greater(*dst, v);
	case ARRAY_UCHAR:
		return reinterpret_cast<const zeUchar*>(ptr)->greater(*dst, v);
	case ARRAY_SHORT:
		return reinterpret_cast<const zeShort*>(ptr)->greater(*dst, v);
	case ARRAY_USHORT:
		return reinterpret_cast<const zeUshort*>(ptr)->greater(*dst, v);
	case ARRAY_INT:
		return reinterpret_cast<const zeInt*>(ptr)->greater(*dst, v);
	case ARRAY_UINT:
		return reinterpret_cast<const zeUint*>(ptr)->greater(*dst, v);
	case ARRAY_FLOAT:
		return reinterpret_cast<const zeFloat*>(ptr)->greater(*dst, v);
	case ARRAY_DOUBLE:
		return reinterpret_cast<const zeDouble*>(ptr)->greater(*dst, v);
	}
	return 0;
}

int
array_greater2(void *idx, const void *ptr1, const void *ptr2)
{
	int type = array_is_int8(idx);
	if (!type) return 0;
	type = array_check_type2(ptr1, ptr2);
	if (!type) return 0;
	zeChar* dst = reinterpret_cast<zeChar*>(idx);
	array_set_error_message("array_greater2() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<const zeChar*>(ptr1)->greater(*dst, *reinterpret_cast<const zeChar*>(ptr2));
	case ARRAY_UCHAR:
		return reinterpret_cast<const zeUchar*>(ptr1)->greater(*dst, *reinterpret_cast<const zeUchar*>(ptr2));
	case ARRAY_SHORT:
		return reinterpret_cast<const zeShort*>(ptr1)->greater(*dst, *reinterpret_cast<const zeShort*>(ptr2));
	case ARRAY_USHORT:
		return reinterpret_cast<const zeUshort*>(ptr1)->greater(*dst, *reinterpret_cast<const zeUshort*>(ptr2));
	case ARRAY_INT:
		return reinterpret_cast<const zeInt*>(ptr1)->greater(*dst, *reinterpret_cast<const zeInt*>(ptr2));
	case ARRAY_UINT:
		return reinterpret_cast<const zeUint*>(ptr1)->greater(*dst, *reinterpret_cast<const zeUint*>(ptr2));
	case ARRAY_FLOAT:
		return reinterpret_cast<const zeFloat*>(ptr1)->greater(*dst, *reinterpret_cast<const zeFloat*>(ptr2));
	case ARRAY_DOUBLE:
		return reinterpret_cast<const zeDouble*>(ptr1)->greater(*dst, *reinterpret_cast<const zeDouble*>(ptr2));
	}
	return 0;
}

int
array_gequal(void *idx, const void *ptr, double v)
{
	int type = array_is_int8(idx);
	if (!type) return 0;
	type = array_check_type(ptr);
	if (!type) return 0;
	zeChar* dst = reinterpret_cast<zeChar*>(idx);
	array_set_error_message("array_gequal() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<const zeChar*>(ptr)->gequal(*dst, v);
	case ARRAY_UCHAR:
		return reinterpret_cast<const zeUchar*>(ptr)->gequal(*dst, v);
	case ARRAY_SHORT:
		return reinterpret_cast<const zeShort*>(ptr)->gequal(*dst, v);
	case ARRAY_USHORT:
		return reinterpret_cast<const zeUshort*>(ptr)->gequal(*dst, v);
	case ARRAY_INT:
		return reinterpret_cast<const zeInt*>(ptr)->gequal(*dst, v);
	case ARRAY_UINT:
		return reinterpret_cast<const zeUint*>(ptr)->gequal(*dst, v);
	case ARRAY_FLOAT:
		return reinterpret_cast<const zeFloat*>(ptr)->gequal(*dst, v);
	case ARRAY_DOUBLE:
		return reinterpret_cast<const zeDouble*>(ptr)->gequal(*dst, v);
	}
	return 0;
}

int
array_gequal2(void *idx, const void *ptr1, const void *ptr2)
{
	int type = array_is_int8(idx);
	if (!type) return 0;
	type = array_check_type2(ptr1, ptr2);
	if (!type) return 0;
	zeChar* dst = reinterpret_cast<zeChar*>(idx);
	array_set_error_message("array_gequal2() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<const zeChar*>(ptr1)->gequal(*dst, *reinterpret_cast<const zeChar*>(ptr2));
	case ARRAY_UCHAR:
		return reinterpret_cast<const zeUchar*>(ptr1)->gequal(*dst, *reinterpret_cast<const zeUchar*>(ptr2));
	case ARRAY_SHORT:
		return reinterpret_cast<const zeShort*>(ptr1)->gequal(*dst, *reinterpret_cast<const zeShort*>(ptr2));
	case ARRAY_USHORT:
		return reinterpret_cast<const zeUshort*>(ptr1)->gequal(*dst, *reinterpret_cast<const zeUshort*>(ptr2));
	case ARRAY_INT:
		return reinterpret_cast<const zeInt*>(ptr1)->gequal(*dst, *reinterpret_cast<const zeInt*>(ptr2));
	case ARRAY_UINT:
		return reinterpret_cast<const zeUint*>(ptr1)->gequal(*dst, *reinterpret_cast<const zeUint*>(ptr2));
	case ARRAY_FLOAT:
		return reinterpret_cast<const zeFloat*>(ptr1)->gequal(*dst, *reinterpret_cast<const zeFloat*>(ptr2));
	case ARRAY_DOUBLE:
		return reinterpret_cast<const zeDouble*>(ptr1)->gequal(*dst, *reinterpret_cast<const zeDouble*>(ptr2));
	}
	return 0;
}

double
array_get_at(const void *ptr, size_t i, size_t j)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	switch (type)
	{
	case ARRAY_CHAR:
		return (*reinterpret_cast<const zeChar*>(ptr))(i, j);
	case ARRAY_UCHAR:
		return (*reinterpret_cast<const zeUchar*>(ptr))(i, j);
	case ARRAY_SHORT:
		return (*reinterpret_cast<const zeShort*>(ptr))(i, j);
	case ARRAY_USHORT:
		return (*reinterpret_cast<const zeUshort*>(ptr))(i, j);
	case ARRAY_INT:
		return (*reinterpret_cast<const zeInt*>(ptr))(i, j);
	case ARRAY_UINT:
		return (*reinterpret_cast<const zeUint*>(ptr))(i, j);
	case ARRAY_FLOAT:
		return (*reinterpret_cast<const zeFloat*>(ptr))(i, j);
	case ARRAY_DOUBLE:
		return (*reinterpret_cast<const zeDouble*>(ptr))(i, j);
	}
	return 0;
}

int
array_set_at(void *ptr, size_t i, size_t j, double v)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	array_set_error_message("array_set_at() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		{
			zeChar *arr = reinterpret_cast<zeChar*>(ptr);
			if (i * j >= arr->ndat()) return 0;
			(*arr)(i, j) = v;
			return 1;
		}
	case ARRAY_UCHAR:
		{
			zeUchar *arr = reinterpret_cast<zeUchar*>(ptr);
			if (i * j >= arr->ndat()) return 0;
			(*arr)(i, j) = v;
			return 1;
		}
	case ARRAY_SHORT:
		{
			zeShort *arr = reinterpret_cast<zeShort*>(ptr);
			if (i * j >= arr->ndat()) return 0;
			(*arr)(i, j) = v;
			return 1;
		}
	case ARRAY_USHORT:
		{
			zeUshort *arr = reinterpret_cast<zeUshort*>(ptr);
			if (i * j >= arr->ndat()) return 0;
			(*arr)(i, j) = v;
			return 1;
		}
	case ARRAY_INT:
		{
			zeInt *arr = reinterpret_cast<zeInt*>(ptr);
			if (i * j >= arr->ndat()) return 0;
			(*arr)(i, j) = v;
			return 1;
		}
	case ARRAY_UINT:
		{
			zeUint *arr = reinterpret_cast<zeUint*>(ptr);
			if (i * j >= arr->ndat()) return 0;
			(*arr)(i, j) = v;
			return 1;
		}
	case ARRAY_FLOAT:
		{
			zeFloat *arr = reinterpret_cast<zeFloat*>(ptr);
			if (i * j >= arr->ndat()) return 0;
			(*arr)(i, j) = v;
			return 1;
		}
	case ARRAY_DOUBLE:
		{
			zeDouble *arr = reinterpret_cast<zeDouble*>(ptr);
			if (i * j >= arr->ndat()) return 0;
			(*arr)(i, j) = v;
			return 1;
		}
	}
	return 0;
}

int
array_get_row(void *dst, const void *src, size_t i)
{
	int type = array_check_type2(dst, src);
	if (!type) return 0;
	array_set_error_message("array_get_row() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<const zeChar*>(src)->getrow(*reinterpret_cast<zeChar*>(dst), i);
	case ARRAY_UCHAR:
		return reinterpret_cast<const zeUchar*>(src)->getrow(*reinterpret_cast<zeUchar*>(dst), i);
	case ARRAY_SHORT:
		return reinterpret_cast<const zeShort*>(src)->getrow(*reinterpret_cast<zeShort*>(dst), i);
	case ARRAY_USHORT:
		return reinterpret_cast<const zeUshort*>(src)->getrow(*reinterpret_cast<zeUshort*>(dst), i);
	case ARRAY_INT:
		return reinterpret_cast<const zeInt*>(src)->getrow(*reinterpret_cast<zeInt*>(dst), i);
	case ARRAY_UINT:
		return reinterpret_cast<const zeUint*>(src)->getrow(*reinterpret_cast<zeUint*>(dst), i);
	case ARRAY_FLOAT:
		return reinterpret_cast<const zeFloat*>(src)->getrow(*reinterpret_cast<zeFloat*>(dst), i);
	case ARRAY_DOUBLE:
		return reinterpret_cast<const zeDouble*>(src)->getrow(*reinterpret_cast<zeDouble*>(dst), i);
	}
	return 0;
}

int
array_get_column(void *dst, const void *src, size_t j)
{
	int type = array_check_type2(dst, src);
	if (!type) return 0;
	array_set_error_message("array_get_column() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<const zeChar*>(src)->getcol(*reinterpret_cast<zeChar*>(dst), j);
	case ARRAY_UCHAR:
		return reinterpret_cast<const zeUchar*>(src)->getcol(*reinterpret_cast<zeUchar*>(dst), j);
	case ARRAY_SHORT:
		return reinterpret_cast<const zeShort*>(src)->getcol(*reinterpret_cast<zeShort*>(dst), j);
	case ARRAY_USHORT:
		return reinterpret_cast<const zeUshort*>(src)->getcol(*reinterpret_cast<zeUshort*>(dst), j);
	case ARRAY_INT:
		return reinterpret_cast<const zeInt*>(src)->getcol(*reinterpret_cast<zeInt*>(dst), j);
	case ARRAY_UINT:
		return reinterpret_cast<const zeUint*>(src)->getcol(*reinterpret_cast<zeUint*>(dst), j);
	case ARRAY_FLOAT:
		return reinterpret_cast<const zeFloat*>(src)->getcol(*reinterpret_cast<zeFloat*>(dst), j);
	case ARRAY_DOUBLE:
		return reinterpret_cast<const zeDouble*>(src)->getcol(*reinterpret_cast<zeDouble*>(dst), j);
	case ARRAY_COMPLEX:
		return reinterpret_cast<const zeComplex*>(src)->getcol(*reinterpret_cast<zeComplex*>(dst), j);
	case ARRAY_TRIPLEX:
		return reinterpret_cast<const zeTriplex*>(src)->getcol(*reinterpret_cast<zeTriplex*>(dst), j);
	case ARRAY_QUADRUPLE:
		return reinterpret_cast<const zeQuadruple*>(src)->getcol(*reinterpret_cast<zeQuadruple*>(dst), j);
	}
	return 0;
}

int
array_set_row(void *dst, double v, size_t i)
{
	int type = array_check_type(dst);
	if (!type) return 0;
	array_set_error_message("array_set_row() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(dst)->setrow(v, i);
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(dst)->setrow(v, i);
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(dst)->setrow(v, i);
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(dst)->setrow(v, i);
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(dst)->setrow(v, i);
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(dst)->setrow(v, i);
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(dst)->setrow(v, i);
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(dst)->setrow(v, i);
	}
	return 0;
}

int
array_set_row2(void *dst, const void *src, size_t i)
{
	int type = array_check_type2(dst, src);
	if (!type) return 0;
	array_set_error_message("array_set_row2() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(dst)->setrow(*reinterpret_cast<const zeChar*>(src), i);
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(dst)->setrow(*reinterpret_cast<const zeUchar*>(src), i);
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(dst)->setrow(*reinterpret_cast<const zeShort*>(src), i);
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(dst)->setrow(*reinterpret_cast<const zeUshort*>(src), i);
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(dst)->setrow(*reinterpret_cast<const zeInt*>(src), i);
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(dst)->setrow(*reinterpret_cast<const zeUint*>(src), i);
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(dst)->setrow(*reinterpret_cast<const zeFloat*>(src), i);
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(dst)->setrow(*reinterpret_cast<const zeDouble*>(src), i);
	case ARRAY_COMPLEX:
		return reinterpret_cast<zeComplex*>(dst)->setrow(*reinterpret_cast<const zeComplex*>(src), i);
	case ARRAY_TRIPLEX:
		return reinterpret_cast<zeTriplex*>(dst)->setrow(*reinterpret_cast<const zeTriplex*>(src), i);
	case ARRAY_QUADRUPLE:
		return reinterpret_cast<zeQuadruple*>(dst)->setrow(*reinterpret_cast<const zeQuadruple*>(src), i);
	}
	return 0;
}

int
array_set_column(void *dst, double v, size_t i)
{
	int type = array_check_type(dst);
	if (!type) return 0;
	array_set_error_message("array_set_column() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(dst)->setcol(v, i);
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(dst)->setcol(v, i);
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(dst)->setcol(v, i);
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(dst)->setcol(v, i);
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(dst)->setcol(v, i);
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(dst)->setcol(v, i);
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(dst)->setcol(v, i);
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(dst)->setcol(v, i);
	}
	return 0;
}

int
array_set_column2(void *dst, const void *src, size_t i)
{
	int type = array_check_type2(dst, src);
	if (!type) return 0;
	array_set_error_message("array_set_column2() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(dst)->setcol(*reinterpret_cast<const zeChar*>(src), i);
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(dst)->setcol(*reinterpret_cast<const zeUchar*>(src), i);
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(dst)->setcol(*reinterpret_cast<const zeShort*>(src), i);
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(dst)->setcol(*reinterpret_cast<const zeUshort*>(src), i);
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(dst)->setcol(*reinterpret_cast<const zeInt*>(src), i);
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(dst)->setcol(*reinterpret_cast<const zeUint*>(src), i);
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(dst)->setcol(*reinterpret_cast<const zeFloat*>(src), i);
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(dst)->setcol(*reinterpret_cast<const zeDouble*>(src), i);
	case ARRAY_COMPLEX:
		return reinterpret_cast<zeComplex*>(dst)->setcol(*reinterpret_cast<const zeComplex*>(src), i);
	case ARRAY_TRIPLEX:
		return reinterpret_cast<zeTriplex*>(dst)->setcol(*reinterpret_cast<const zeTriplex*>(src), i);
	case ARRAY_QUADRUPLE:
		return reinterpret_cast<zeQuadruple*>(dst)->setcol(*reinterpret_cast<const zeQuadruple*>(src), i);
	}
	return 0;
}

int
array_insert_row(void *dst, double v, size_t i)
{
	int type = array_check_type(dst);
	if (!type) return 0;
	array_set_error_message("array_insert_row() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(dst)->insertrow(v, i);
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(dst)->insertrow(v, i);
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(dst)->insertrow(v, i);
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(dst)->insertrow(v, i);
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(dst)->insertrow(v, i);
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(dst)->insertrow(v, i);
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(dst)->insertrow(v, i);
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(dst)->insertrow(v, i);
	}
	return 0;
}

int
array_insert_row2(void *dst, const void *src, size_t i)
{
	int type = array_check_type2(dst, src);
	if (!type) return 0;
	array_set_error_message("array_insert_row2() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(dst)->insertrow(*reinterpret_cast<const zeChar*>(src), i);
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(dst)->insertrow(*reinterpret_cast<const zeUchar*>(src), i);
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(dst)->insertrow(*reinterpret_cast<const zeShort*>(src), i);
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(dst)->insertrow(*reinterpret_cast<const zeUshort*>(src), i);
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(dst)->insertrow(*reinterpret_cast<const zeInt*>(src), i);
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(dst)->insertrow(*reinterpret_cast<const zeUint*>(src), i);
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(dst)->insertrow(*reinterpret_cast<const zeFloat*>(src), i);
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(dst)->insertrow(*reinterpret_cast<const zeDouble*>(src), i);
	case ARRAY_COMPLEX:
		return reinterpret_cast<zeComplex*>(dst)->insertrow(*reinterpret_cast<const zeComplex*>(src), i);
	case ARRAY_TRIPLEX:
		return reinterpret_cast<zeTriplex*>(dst)->insertrow(*reinterpret_cast<const zeTriplex*>(src), i);
	case ARRAY_QUADRUPLE:
		return reinterpret_cast<zeQuadruple*>(dst)->insertrow(*reinterpret_cast<const zeQuadruple*>(src), i);
	}
	return 0;
}

int
array_insert_column(void *dst, double v, size_t i)
{
	int type = array_check_type(dst);
	if (!type) return 0;
	array_set_error_message("array_insert_column() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(dst)->insertcol(v, i);
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(dst)->insertcol(v, i);
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(dst)->insertcol(v, i);
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(dst)->insertcol(v, i);
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(dst)->insertcol(v, i);
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(dst)->insertcol(v, i);
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(dst)->insertcol(v, i);
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(dst)->insertcol(v, i);
	}
	return 0;
}

int
array_insert_column2(void *dst, const void *src, size_t i)
{
	int type = array_check_type2(dst, src);
	if (!type) return 0;
	array_set_error_message("array_insert_column2() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(dst)->insertcol(*reinterpret_cast<const zeChar*>(src), i);
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(dst)->insertcol(*reinterpret_cast<const zeUchar*>(src), i);
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(dst)->insertcol(*reinterpret_cast<const zeShort*>(src), i);
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(dst)->insertcol(*reinterpret_cast<const zeUshort*>(src), i);
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(dst)->insertcol(*reinterpret_cast<const zeInt*>(src), i);
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(dst)->insertcol(*reinterpret_cast<const zeUint*>(src), i);
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(dst)->insertcol(*reinterpret_cast<const zeFloat*>(src), i);
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(dst)->insertcol(*reinterpret_cast<const zeDouble*>(src), i);
	case ARRAY_COMPLEX:
		return reinterpret_cast<zeComplex*>(dst)->insertcol(*reinterpret_cast<const zeComplex*>(src), i);
	case ARRAY_TRIPLEX:
		return reinterpret_cast<zeTriplex*>(dst)->insertcol(*reinterpret_cast<const zeTriplex*>(src), i);
	case ARRAY_QUADRUPLE:
		return reinterpret_cast<zeQuadruple*>(dst)->insertcol(*reinterpret_cast<const zeQuadruple*>(src), i);
	}
	return 0;
}

int
array_delete_row(void *ptr, size_t i)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	array_set_error_message("array_delete_row() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(ptr)->delrow(i);
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(ptr)->delrow(i);
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(ptr)->delrow(i);
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(ptr)->delrow(i);
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(ptr)->delrow(i);
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(ptr)->delrow(i);
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(ptr)->delrow(i);
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(ptr)->delrow(i);
	case ARRAY_COMPLEX:
		return reinterpret_cast<zeComplex*>(ptr)->delrow(i);
	case ARRAY_TRIPLEX:
		return reinterpret_cast<zeTriplex*>(ptr)->delrow(i);
	case ARRAY_QUADRUPLE:
		return reinterpret_cast<zeQuadruple*>(ptr)->delrow(i);
	}
	return 0;
}

int
array_delete_column(void *ptr, size_t i)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	array_set_error_message("array_delete_column() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(ptr)->delcol(i);
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(ptr)->delcol(i);
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(ptr)->delcol(i);
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(ptr)->delcol(i);
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(ptr)->delcol(i);
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(ptr)->delcol(i);
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(ptr)->delcol(i);
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(ptr)->delcol(i);
	case ARRAY_COMPLEX:
		return reinterpret_cast<zeComplex*>(ptr)->delcol(i);
	case ARRAY_TRIPLEX:
		return reinterpret_cast<zeTriplex*>(ptr)->delcol(i);
	case ARRAY_QUADRUPLE:
		return reinterpret_cast<zeQuadruple*>(ptr)->delcol(i);
	}
	return 0;
}

int
array_get_by_index(void *dst, const void *src, const void *idx)
{
	int type = array_is_int8(idx);
	if (!type) return 0;
	type = array_check_type2(dst, src);
	if (!type) return 0;
	const zeChar *arr = reinterpret_cast<const zeChar*>(idx);
	array_set_error_message("array_get_by_index() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<const zeChar*>(src)->getidx(*reinterpret_cast<zeChar*>(dst), *arr);
	case ARRAY_UCHAR:
		return reinterpret_cast<const zeUchar*>(src)->getidx(*reinterpret_cast<zeUchar*>(dst), *arr);
	case ARRAY_SHORT:
		return reinterpret_cast<const zeShort*>(src)->getidx(*reinterpret_cast<zeShort*>(dst), *arr);
	case ARRAY_USHORT:
		return reinterpret_cast<const zeUshort*>(src)->getidx(*reinterpret_cast<zeUshort*>(dst), *arr);
	case ARRAY_INT:
		return reinterpret_cast<const zeInt*>(src)->getidx(*reinterpret_cast<zeInt*>(dst), *arr);
	case ARRAY_UINT:
		return reinterpret_cast<const zeUint*>(src)->getidx(*reinterpret_cast<zeUint*>(dst), *arr);
	case ARRAY_FLOAT:
		return reinterpret_cast<const zeFloat*>(src)->getidx(*reinterpret_cast<zeFloat*>(dst), *arr);
	case ARRAY_DOUBLE:
		return reinterpret_cast<const zeDouble*>(src)->getidx(*reinterpret_cast<zeDouble*>(dst), *arr);
	}
	return 0;
}

int
array_set_by_index(void *dst, double v, const void *idx)
{
	int type = array_is_int8(idx);
	if (!type) return 0;
	type = array_check_type(dst);
	if (!type) return 0;
	const zeChar *arr = reinterpret_cast<const zeChar*>(idx);
	array_set_error_message("array_set_by_index() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(dst)->setidx(*arr, v);
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(dst)->setidx(*arr, v);
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(dst)->setidx(*arr, v);
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(dst)->setidx(*arr, v);
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(dst)->setidx(*arr, v);
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(dst)->setidx(*arr, v);
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(dst)->setidx(*arr, v);
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(dst)->setidx(*arr, v);
	}
	return 0;
}

int
array_set_by_index2(void *dst, const void *src, const void *idx)
{
	int type = array_is_int8(idx);
	if (!type) return 0;
	type = array_check_type2(dst, src);
	if (!type) return 0;
	const zeChar *arr = reinterpret_cast<const zeChar*>(idx);
	array_set_error_message("array_set_by_index2() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(dst)->setidx(*arr, *reinterpret_cast<const zeChar*>(src));
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(dst)->setidx(*arr, *reinterpret_cast<const zeUchar*>(src));
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(dst)->setidx(*arr, *reinterpret_cast<const zeShort*>(src));
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(dst)->setidx(*arr, *reinterpret_cast<const zeUshort*>(src));
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(dst)->setidx(*arr, *reinterpret_cast<const zeInt*>(src));
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(dst)->setidx(*arr, *reinterpret_cast<const zeUint*>(src));
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(dst)->setidx(*arr, *reinterpret_cast<const zeFloat*>(src));
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(dst)->setidx(*arr, *reinterpret_cast<const zeDouble*>(src));
	}
	return 0;
}

int
array_get_by_block(void *dst, const void *src, size_t i, size_t m, size_t j, size_t n)
{
	int type = array_check_type2(dst, src);
	if (!type) return 0;
	array_set_error_message("array_get_by_block() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<const zeChar*>(src)->getblc(*reinterpret_cast<zeChar*>(dst), i, m, j, n);
	case ARRAY_UCHAR:
		return reinterpret_cast<const zeUchar*>(src)->getblc(*reinterpret_cast<zeUchar*>(dst), i, m, j, n);
	case ARRAY_SHORT:
		return reinterpret_cast<const zeShort*>(src)->getblc(*reinterpret_cast<zeShort*>(dst), i, m, j, n);
	case ARRAY_USHORT:
		return reinterpret_cast<const zeUshort*>(src)->getblc(*reinterpret_cast<zeUshort*>(dst), i, m, j, n);
	case ARRAY_INT:
		return reinterpret_cast<const zeInt*>(src)->getblc(*reinterpret_cast<zeInt*>(dst), i, m, j, n);
	case ARRAY_UINT:
		return reinterpret_cast<const zeUint*>(src)->getblc(*reinterpret_cast<zeUint*>(dst), i, m, j, n);
	case ARRAY_FLOAT:
		return reinterpret_cast<const zeFloat*>(src)->getblc(*reinterpret_cast<zeFloat*>(dst), i, m, j, n);
	case ARRAY_DOUBLE:
		return reinterpret_cast<const zeDouble*>(src)->getblc(*reinterpret_cast<zeDouble*>(dst), i, m, j, n);
	}
	return 0;
}

int
array_set_by_block(void *dst, const void *src, size_t i, size_t j)
{
	int type = array_check_type2(dst, src);
	if (!type) return 0;
	array_set_error_message("array_set_by_block() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(dst)->setblc(*reinterpret_cast<const zeChar*>(src), i, j);
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(dst)->setblc(*reinterpret_cast<const zeUchar*>(src), i, j);
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(dst)->setblc(*reinterpret_cast<const zeShort*>(src), i, j);
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(dst)->setblc(*reinterpret_cast<const zeUshort*>(src), i, j);
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(dst)->setblc(*reinterpret_cast<const zeInt*>(src), i, j);
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(dst)->setblc(*reinterpret_cast<const zeUint*>(src), i, j);
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(dst)->setblc(*reinterpret_cast<const zeFloat*>(src), i, j);
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(dst)->setblc(*reinterpret_cast<const zeDouble*>(src), i, j);
	}
	return 0;
}

int
array_find_eq(void *idx, const void *ptr, double v)
{
	int type = array_is_uint32(idx);
	if (!type) return 0;
	type = array_check_type(ptr);
	if (!type) return 0;
	zeUint *arr = reinterpret_cast<zeUint*>(idx);
	array_set_error_message("array_find() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<const zeChar*>(ptr)->findeq(*arr, v);
	case ARRAY_UCHAR:
		return reinterpret_cast<const zeUchar*>(ptr)->findeq(*arr, v);
	case ARRAY_SHORT:
		return reinterpret_cast<const zeShort*>(ptr)->findeq(*arr, v);
	case ARRAY_USHORT:
		return reinterpret_cast<const zeUshort*>(ptr)->findeq(*arr, v);
	case ARRAY_INT:
		return reinterpret_cast<const zeInt*>(ptr)->findeq(*arr, v);
	case ARRAY_UINT:
		return reinterpret_cast<const zeUint*>(ptr)->findeq(*arr, v);
	case ARRAY_FLOAT:
		return reinterpret_cast<const zeFloat*>(ptr)->findeq(*arr, v);
	case ARRAY_DOUBLE:
		return reinterpret_cast<const zeDouble*>(ptr)->findeq(*arr, v);
	}
	return 0;
}

int
array_find_ne(void *idx, const void *ptr, double v)
{
	int type = array_is_uint32(idx);
	if (!type) return 0;
	type = array_check_type(ptr);
	if (!type) return 0;
	zeUint *arr = reinterpret_cast<zeUint*>(idx);
	array_set_error_message("array_find() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<const zeChar*>(ptr)->findne(*arr, v);
	case ARRAY_UCHAR:
		return reinterpret_cast<const zeUchar*>(ptr)->findne(*arr, v);
	case ARRAY_SHORT:
		return reinterpret_cast<const zeShort*>(ptr)->findne(*arr, v);
	case ARRAY_USHORT:
		return reinterpret_cast<const zeUshort*>(ptr)->findne(*arr, v);
	case ARRAY_INT:
		return reinterpret_cast<const zeInt*>(ptr)->findne(*arr, v);
	case ARRAY_UINT:
		return reinterpret_cast<const zeUint*>(ptr)->findne(*arr, v);
	case ARRAY_FLOAT:
		return reinterpret_cast<const zeFloat*>(ptr)->findne(*arr, v);
	case ARRAY_DOUBLE:
		return reinterpret_cast<const zeDouble*>(ptr)->findne(*arr, v);
	}
	return 0;
}

int
array_transpose(void *ptr)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	array_set_error_message("array_transpose() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(ptr)->transpose();
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(ptr)->transpose();
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(ptr)->transpose();
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(ptr)->transpose();
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(ptr)->transpose();
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(ptr)->transpose();
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(ptr)->transpose();
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(ptr)->transpose();
	}
	return 0;
}

int
array_swap_rows(void *ptr, size_t i, size_t j)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	array_set_error_message("array_swap_column() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(ptr)->swaprow(i, j);
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(ptr)->swaprow(i, j);
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(ptr)->swaprow(i, j);
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(ptr)->swaprow(i, j);
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(ptr)->swaprow(i, j);
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(ptr)->swaprow(i, j);
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(ptr)->swaprow(i, j);
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(ptr)->swaprow(i, j);
	case ARRAY_COMPLEX:
		return reinterpret_cast<zeComplex*>(ptr)->swaprow(i, j);
	case ARRAY_TRIPLEX:
		return reinterpret_cast<zeTriplex*>(ptr)->swaprow(i, j);
	case ARRAY_QUADRUPLE:
		return reinterpret_cast<zeQuadruple*>(ptr)->swaprow(i, j);
	}
	return 0;
}

int
array_swap_columns(void *ptr, size_t i, size_t j)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	array_set_error_message("array_swap_column() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(ptr)->swapcol(i, j);
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(ptr)->swapcol(i, j);
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(ptr)->swapcol(i, j);
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(ptr)->swapcol(i, j);
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(ptr)->swapcol(i, j);
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(ptr)->swapcol(i, j);
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(ptr)->swapcol(i, j);
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(ptr)->swapcol(i, j);
	case ARRAY_COMPLEX:
		return reinterpret_cast<zeComplex*>(ptr)->swapcol(i, j);
	case ARRAY_TRIPLEX:
		return reinterpret_cast<zeTriplex*>(ptr)->swapcol(i, j);
	case ARRAY_QUADRUPLE:
		return reinterpret_cast<zeQuadruple*>(ptr)->swapcol(i, j);
	}
	return 0;
}

int
array_flip_rows(void *ptr)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	array_set_error_message("array_flip_rows() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(ptr)->fliprow();
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(ptr)->fliprow();
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(ptr)->fliprow();
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(ptr)->fliprow();
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(ptr)->fliprow();
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(ptr)->fliprow();
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(ptr)->fliprow();
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(ptr)->fliprow();
	case ARRAY_COMPLEX:
		return reinterpret_cast<zeComplex*>(ptr)->fliprow();
	case ARRAY_TRIPLEX:
		return reinterpret_cast<zeTriplex*>(ptr)->fliprow();
	case ARRAY_QUADRUPLE:
		return reinterpret_cast<zeQuadruple*>(ptr)->fliprow();
	}
	return 0;
}

int
array_flip_columns(void *ptr)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	array_set_error_message("array_flip_rows() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(ptr)->flipcol();
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(ptr)->flipcol();
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(ptr)->flipcol();
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(ptr)->flipcol();
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(ptr)->flipcol();
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(ptr)->flipcol();
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(ptr)->flipcol();
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(ptr)->flipcol();
	case ARRAY_COMPLEX:
		return reinterpret_cast<zeComplex*>(ptr)->flipcol();
	case ARRAY_TRIPLEX:
		return reinterpret_cast<zeTriplex*>(ptr)->flipcol();
	case ARRAY_QUADRUPLE:
		return reinterpret_cast<zeQuadruple*>(ptr)->flipcol();
	}
	return 0;
}

int
array_shift_rows(void *ptr, int n)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	array_set_error_message("array_shift_rows() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(ptr)->shiftrow(n);
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(ptr)->shiftrow(n);
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(ptr)->shiftrow(n);
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(ptr)->shiftrow(n);
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(ptr)->shiftrow(n);
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(ptr)->shiftrow(n);
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(ptr)->shiftrow(n);
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(ptr)->shiftrow(n);
	case ARRAY_COMPLEX:
		return reinterpret_cast<zeComplex*>(ptr)->shiftrow(n);
	case ARRAY_TRIPLEX:
		return reinterpret_cast<zeTriplex*>(ptr)->shiftrow(n);
	case ARRAY_QUADRUPLE:
		return reinterpret_cast<zeQuadruple*>(ptr)->shiftrow(n);
	}
	return 0;
}

int
array_shift_columns(void *ptr, int n)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	array_set_error_message("array_shift_columns() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(ptr)->shiftcol(n);
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(ptr)->shiftcol(n);
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(ptr)->shiftcol(n);
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(ptr)->shiftcol(n);
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(ptr)->shiftcol(n);
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(ptr)->shiftcol(n);
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(ptr)->shiftcol(n);
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(ptr)->shiftcol(n);
	case ARRAY_COMPLEX:
		return reinterpret_cast<zeComplex*>(ptr)->shiftcol(n);
	case ARRAY_TRIPLEX:
		return reinterpret_cast<zeTriplex*>(ptr)->shiftcol(n);
	case ARRAY_QUADRUPLE:
		return reinterpret_cast<zeQuadruple*>(ptr)->shiftcol(n);
	}
	return 0;
}

int
array_product(void *dst, const void *src)
{
	int type = array_check_type2(dst, src);
	if (!type) return 0;
	array_set_error_message("array_product() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(dst)->concat(*reinterpret_cast<const zeChar*>(src));
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(dst)->concat(*reinterpret_cast<const zeUchar*>(src));
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(dst)->concat(*reinterpret_cast<const zeShort*>(src));
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(dst)->concat(*reinterpret_cast<const zeUshort*>(src));
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(dst)->concat(*reinterpret_cast<const zeInt*>(src));
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(dst)->concat(*reinterpret_cast<const zeUint*>(src));
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(dst)->concat(*reinterpret_cast<const zeFloat*>(src));
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(dst)->concat(*reinterpret_cast<const zeDouble*>(src));
	}
	return 0;
}

int
array_sort(void *ptr, size_t i, int assending)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	array_set_error_message("array_sort() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(ptr)->sort(i, assending);
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(ptr)->sort(i, assending);
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(ptr)->sort(i, assending);
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(ptr)->sort(i, assending);
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(ptr)->sort(i, assending);
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(ptr)->sort(i, assending);
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(ptr)->sort(i, assending);
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(ptr)->sort(i, assending);
	}
	return 0;
}

int
array_unique(void *ptr, int assending)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	array_set_error_message("array_unique() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(ptr)->unique(assending);
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(ptr)->unique(assending);
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(ptr)->unique(assending);
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(ptr)->unique(assending);
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(ptr)->unique(assending);
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(ptr)->unique(assending);
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(ptr)->unique(assending);
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(ptr)->unique(assending);
	}
	return 0;
}

double
array_min(const void *ptr)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	array_set_error_message("array_min() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<const zeChar*>(ptr)->vmin(127);
	case ARRAY_UCHAR:
		return reinterpret_cast<const zeUchar*>(ptr)->vmin(255);
	case ARRAY_SHORT:
		return reinterpret_cast<const zeShort*>(ptr)->vmin(32767);
	case ARRAY_USHORT:
		return reinterpret_cast<const zeUshort*>(ptr)->vmin(65535);
	case ARRAY_INT:
		return reinterpret_cast<const zeInt*>(ptr)->vmin(2147483647);
	case ARRAY_UINT:
		return reinterpret_cast<const zeUint*>(ptr)->vmin(4294967295);
	case ARRAY_FLOAT:
		return reinterpret_cast<const zeFloat*>(ptr)->vmin((float)1.e30);
	case ARRAY_DOUBLE:
		return reinterpret_cast<const zeDouble*>(ptr)->vmin(1.e30);
	}
	return 0;
}

double
array_max(const void *ptr)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	array_set_error_message("array_max() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<const zeChar*>(ptr)->vmax(-127);
	case ARRAY_UCHAR:
		return reinterpret_cast<const zeUchar*>(ptr)->vmax(0);
	case ARRAY_SHORT:
		return reinterpret_cast<const zeShort*>(ptr)->vmax(-32767);
	case ARRAY_USHORT:
		return reinterpret_cast<const zeUshort*>(ptr)->vmax(0);
	case ARRAY_INT:
		return reinterpret_cast<const zeInt*>(ptr)->vmax(-2147483647);
	case ARRAY_UINT:
		return reinterpret_cast<const zeUint*>(ptr)->vmax(0);
	case ARRAY_FLOAT:
		return reinterpret_cast<const zeFloat*>(ptr)->vmax((float)-1.e30);
	case ARRAY_DOUBLE:
		return reinterpret_cast<const zeDouble*>(ptr)->vmax(-1.e30);
	}
	return 0;
}

double
array_sum(const void *ptr)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	array_set_error_message("array_max() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<const zeChar*>(ptr)->sum();
	case ARRAY_UCHAR:
		return reinterpret_cast<const zeUchar*>(ptr)->sum();
	case ARRAY_SHORT:
		return reinterpret_cast<const zeShort*>(ptr)->sum();
	case ARRAY_USHORT:
		return reinterpret_cast<const zeUshort*>(ptr)->sum();
	case ARRAY_INT:
		return reinterpret_cast<const zeInt*>(ptr)->sum();
	case ARRAY_UINT:
		return reinterpret_cast<const zeUint*>(ptr)->sum();
	case ARRAY_FLOAT:
		return reinterpret_cast<const zeFloat*>(ptr)->sum();
	case ARRAY_DOUBLE:
		return reinterpret_cast<const zeDouble*>(ptr)->sum();
	}
	return 0;
}

int
array_confine(void *ptr, double vmin, double vmax)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	array_set_error_message("array_max() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(ptr)->confine(vmin, vmax);
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(ptr)->confine(vmin, vmax);
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(ptr)->confine(vmin, vmax);
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(ptr)->confine(vmin, vmax);
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(ptr)->confine(vmin, vmax);
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(ptr)->confine(vmin, vmax);
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(ptr)->confine(vmin, vmax);
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(ptr)->confine(vmin, vmax);
	}
	return 0;
}

int
array_print(const void *ptr, const char *format, FILE *fp)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	array_set_error_message("array_print() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		if (!format) format = "  %d";
		return reinterpret_cast<const zeChar*>(ptr)->print(format, fp);
	case ARRAY_UCHAR:
		if (!format) format = "  %u";
		return reinterpret_cast<const zeUchar*>(ptr)->print(format, fp);
	case ARRAY_SHORT:
		if (!format) format = "  %d";
		return reinterpret_cast<const zeShort*>(ptr)->print(format, fp);
	case ARRAY_USHORT:
		if (!format) format = "  %u";
		return reinterpret_cast<const zeUshort*>(ptr)->print(format, fp);
	case ARRAY_INT:
		if (!format) format = "  %d";
		return reinterpret_cast<const zeInt*>(ptr)->print(format, fp);
	case ARRAY_UINT:
		if (!format) format = "  %u";
		return reinterpret_cast<const zeUint*>(ptr)->print(format, fp);
	case ARRAY_FLOAT:
		if (!format) format = "  %#6f";
		return reinterpret_cast<const zeFloat*>(ptr)->print(format, fp);
	case ARRAY_DOUBLE:
		if (!format) format = "  %#6f";
		return reinterpret_cast<const zeDouble*>(ptr)->print(format, fp);
	case ARRAY_COMPLEX:
		{
		if (!format) format = " (%#6f, %#6f)";
		const zeComplex *arr = reinterpret_cast<const zeComplex*>(ptr);
		for (int i = 0; i < arr->nrow(); i++) {
			for (int j = 0; j < arr->ncol(); j++) {
				sComplex v = (*arr)(i, j);
				fprintf(fp, format, v.r, v.i);
			}
			fprintf(fp, "\n");
		}
		return 1;
		}
	case ARRAY_TRIPLEX:
		{
		if (!format) format = " (%#6f, %#6f, %#6f)";
		const zeTriplex *arr = reinterpret_cast<const zeTriplex*>(ptr);
		for (int i = 0; i < arr->nrow(); i++) {
			for (int j = 0; j < arr->ncol(); j++) {
				sTriplex v = (*arr)(i, j);
				fprintf(fp, format, v.x, v.y, v.z);
			}
			fprintf(fp, "\n");
		}
		return 1;
		}
	case ARRAY_QUADRUPLE:
		{
		if (!format) format = " (%#6f, %#6f, %#6f, %#6f)";
		const zeQuadruple *arr = reinterpret_cast<const zeQuadruple*>(ptr);
		for (int i = 0; i < arr->nrow(); i++) {
			for (int j = 0; j < arr->ncol(); j++) {
				sQuadruple v = (*arr)(i, j);
				fprintf(fp, format, v.x, v.y, v.z, v.w);
			}
			fprintf(fp, "\n");
		}
		return 1;
		}
	}
	return 0;
}

int
array_clone(void *dst, const void *src)
{
	int type_dst = array_check_type(dst);
	int type_src = array_check_type(src);
	if (!type_dst || !type_src) return 0;
	array_set_error_message("array_clone() error.");
	size_t i, n, nrow, ncol;
	switch (type_dst)
	{
	case ARRAY_CHAR:
		{
			zeChar* me = reinterpret_cast<zeChar*>(dst);
			switch (type_src)
			{
			case ARRAY_CHAR:
				*me = *(reinterpret_cast<const zeChar*>(src));
				return 1;
			case ARRAY_UCHAR:
				{
					const zeUchar* arr = reinterpret_cast<const zeUchar*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					memcpy(me->data(), arr->data(), nrow*ncol);
				}
				return 1;
			case ARRAY_SHORT:
				{
					const zeShort* arr = reinterpret_cast<const zeShort*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_USHORT:
				{
					const zeUshort* arr = reinterpret_cast<const zeUshort*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_INT:
				{
					const zeInt* arr = reinterpret_cast<const zeInt*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_UINT:
				{
					const zeUint* arr = reinterpret_cast<const zeUint*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_FLOAT:
				{
					const zeFloat* arr = reinterpret_cast<const zeFloat*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_DOUBLE:
				{
					const zeDouble* arr = reinterpret_cast<const zeDouble*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			}
		}
		return 0;
	case ARRAY_UCHAR:
		{
			zeUchar* me = reinterpret_cast<zeUchar*>(dst);
			switch (type_src)
			{
			case ARRAY_CHAR:
				{
					const zeChar* arr = reinterpret_cast<const zeChar*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					memcpy(me->data(), arr->data(), nrow*ncol);
				}
				return 1;
			case ARRAY_UCHAR:
				*me = *(reinterpret_cast<const zeUchar*>(src));
				return 1;
			case ARRAY_SHORT:
				{
					const zeShort* arr = reinterpret_cast<const zeShort*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_USHORT:
				{
					const zeUshort* arr = reinterpret_cast<const zeUshort*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_INT:
				{
					const zeInt* arr = reinterpret_cast<const zeInt*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_UINT:
				{
					const zeUint* arr = reinterpret_cast<const zeUint*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_FLOAT:
				{
					const zeFloat* arr = reinterpret_cast<const zeFloat*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_DOUBLE:
				{
					const zeDouble* arr = reinterpret_cast<const zeDouble*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			}
		}
		return 0;
	case ARRAY_SHORT:
		{
			zeShort* me = reinterpret_cast<zeShort*>(dst);
			switch (type_src)
			{
			case ARRAY_CHAR:
				{
					const zeChar* arr = reinterpret_cast<const zeChar*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_UCHAR:
				{
					const zeUchar* arr = reinterpret_cast<const zeUchar*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_SHORT:
				*me = *(reinterpret_cast<const zeShort*>(src));
				return 1;
			case ARRAY_USHORT:
				{
					const zeUshort* arr = reinterpret_cast<const zeUshort*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					memcpy(me->data(), arr->data(), sizeof(short)*nrow*ncol);
				}
				return 1;
			case ARRAY_INT:
				{
					const zeInt* arr = reinterpret_cast<const zeInt*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_UINT:
				{
					const zeUint* arr = reinterpret_cast<const zeUint*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_FLOAT:
				{
					const zeFloat* arr = reinterpret_cast<const zeFloat*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_DOUBLE:
				{
					const zeDouble* arr = reinterpret_cast<const zeDouble*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			}
		}
		return 0;
	case ARRAY_USHORT:
		{
			zeUshort* me = reinterpret_cast<zeUshort*>(dst);
			switch (type_src)
			{
			case ARRAY_CHAR:
				{
					const zeChar* arr = reinterpret_cast<const zeChar*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_UCHAR:
				{
					const zeUchar* arr = reinterpret_cast<const zeUchar*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_SHORT:
				{
					const zeShort* arr = reinterpret_cast<const zeShort*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					memcpy(me->data(), arr->data(), sizeof(short)*nrow*ncol);
				}
				return 1;
			case ARRAY_USHORT:
				*me = *(reinterpret_cast<const zeUshort*>(src));
				return 1;
			case ARRAY_INT:
				{
					const zeInt* arr = reinterpret_cast<const zeInt*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_UINT:
				{
					const zeUint* arr = reinterpret_cast<const zeUint*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_FLOAT:
				{
					const zeFloat* arr = reinterpret_cast<const zeFloat*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_DOUBLE:
				{
					const zeDouble* arr = reinterpret_cast<const zeDouble*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			}
		}
		return 0;
	case ARRAY_INT:
		{
			zeInt* me = reinterpret_cast<zeInt*>(dst);
			switch (type_src)
			{
			case ARRAY_CHAR:
				{
					const zeChar* arr = reinterpret_cast<const zeChar*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_UCHAR:
				{
					const zeUchar* arr = reinterpret_cast<const zeUchar*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_SHORT:
				{
					const zeShort* arr = reinterpret_cast<const zeShort*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_USHORT:
				{
					const zeUshort* arr = reinterpret_cast<const zeUshort*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_INT:
				*me = *(reinterpret_cast<const zeInt*>(src));
				return 1;
			case ARRAY_UINT:
				{
					const zeUint* arr = reinterpret_cast<const zeUint*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					memcpy(me->data(), arr->data(), sizeof(int)*nrow*ncol);
				}
				return 1;
			case ARRAY_FLOAT:
				{
					const zeFloat* arr = reinterpret_cast<const zeFloat*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_DOUBLE:
				{
					const zeDouble* arr = reinterpret_cast<const zeDouble*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			}
		}
		return 0;
	case ARRAY_UINT:
		{
			zeUint* me = reinterpret_cast<zeUint*>(dst);
			switch (type_src)
			{
			case ARRAY_CHAR:
				{
					const zeChar* arr = reinterpret_cast<const zeChar*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_UCHAR:
				{
					const zeUchar* arr = reinterpret_cast<const zeUchar*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_SHORT:
				{
					const zeShort* arr = reinterpret_cast<const zeShort*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_USHORT:
				{
					const zeUshort* arr = reinterpret_cast<const zeUshort*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_INT:
				{
					const zeInt* arr = reinterpret_cast<const zeInt*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					memcpy(me->data(), arr->data(), sizeof(int)*nrow*ncol);
				}
				return 1;
			case ARRAY_UINT:
				*me = *(reinterpret_cast<const zeUint*>(src));
				return 1;
			case ARRAY_FLOAT:
				{
					const zeFloat* arr = reinterpret_cast<const zeFloat*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_DOUBLE:
				{
					const zeDouble* arr = reinterpret_cast<const zeDouble*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			}
		}
		return 0;
	case ARRAY_FLOAT:
		{
			zeFloat* me = reinterpret_cast<zeFloat*>(dst);
			switch (type_src)
			{
			case ARRAY_CHAR:
				{
					const zeChar* arr = reinterpret_cast<const zeChar*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_UCHAR:
				{
					const zeUchar* arr = reinterpret_cast<const zeUchar*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_SHORT:
				{
					const zeShort* arr = reinterpret_cast<const zeShort*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_USHORT:
				{
					const zeUshort* arr = reinterpret_cast<const zeUshort*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_INT:
				{
					const zeInt* arr = reinterpret_cast<const zeInt*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_UINT:
				{
					const zeUint* arr = reinterpret_cast<const zeUint*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_FLOAT:
				*me = *(reinterpret_cast<const zeFloat*>(src));
				return 1;
			case ARRAY_DOUBLE:
				{
					const zeDouble* arr = reinterpret_cast<const zeDouble*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			}
		}
		return 0;
	case ARRAY_DOUBLE:
		{
			zeDouble* me = reinterpret_cast<zeDouble*>(dst);
			switch (type_src)
			{
			case ARRAY_CHAR:
				{
					const zeChar* arr = reinterpret_cast<const zeChar*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_UCHAR:
				{
					const zeUchar* arr = reinterpret_cast<const zeUchar*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_SHORT:
				{
					const zeShort* arr = reinterpret_cast<const zeShort*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_USHORT:
				{
					const zeUshort* arr = reinterpret_cast<const zeUshort*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_INT:
				{
					const zeInt* arr = reinterpret_cast<const zeInt*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_UINT:
				{
					const zeUint* arr = reinterpret_cast<const zeUint*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_FLOAT:
				{
					const zeFloat* arr = reinterpret_cast<const zeFloat*>(src);
					nrow = arr->nrow();
					ncol = arr->ncol();
					me->resize(nrow, ncol);
					n = nrow * ncol;
					for (i = 0; i < n; i++) (*me)(i) = (*arr)(i);
				}
				return 1;
			case ARRAY_DOUBLE:
				*me = *(reinterpret_cast<const zeDouble*>(src));
				return 1;
			case ARRAY_COMPLEX:
				{
				zeDouble *d = reinterpret_cast<zeDouble*>(dst);
				const zeComplex *s = reinterpret_cast<const zeComplex*>(src);
				d->resize(s->ndat(), 2);
				memcpy(d->data(), s->data(), d->ndat()*sizeof(double));
				}
				return 1;
			case ARRAY_TRIPLEX:
				{
				zeDouble *d = reinterpret_cast<zeDouble*>(dst);
				const zeTriplex *s = reinterpret_cast<const zeTriplex*>(src);
				d->resize(s->ndat(), 3);
				memcpy(d->data(), s->data(), d->ndat()*sizeof(double));
				}
				return 1;
			case ARRAY_QUADRUPLE:
				{
				zeDouble *d = reinterpret_cast<zeDouble*>(dst);
				const zeQuadruple *s = reinterpret_cast<const zeQuadruple*>(src);
				d->resize(s->ndat(), 4);
				memcpy(d->data(), s->data(), d->ndat()*sizeof(double));
				}
				return 1;
			}
		}
		return 0;
	case ARRAY_COMPLEX:
		if (type_src == ARRAY_COMPLEX) {
			*(reinterpret_cast<zeComplex*>(dst)) = *(reinterpret_cast<const zeComplex*>(src));
			return 1;
		}
		if (type_src == ARRAY_DOUBLE) {
			const zeDouble *s = reinterpret_cast<const zeDouble*>(src);
			zeComplex *d = reinterpret_cast<zeComplex*>(dst);
			if (s->ndat() < 2) return 0;
			d->resize(s->ndat() / 2, 1);
			memcpy(d->data(), s->data(), d->ndat()*sizeof(sComplex));
			return 1;
		}
		return 0;
	case ARRAY_TRIPLEX:
		if (type_src == ARRAY_TRIPLEX) {
			*(reinterpret_cast<zeTriplex*>(dst)) = *(reinterpret_cast<const zeTriplex*>(src));
			return 1;
		}
		if (type_src == ARRAY_DOUBLE) {
			const zeDouble *s = reinterpret_cast<const zeDouble*>(src);
			zeTriplex *d = reinterpret_cast<zeTriplex*>(dst);
			if (s->ndat() < 3) return 0;
			d->resize(s->ndat() / 3, 1);
			memcpy(d->data(), s->data(), d->ndat()*sizeof(sTriplex));
			return 1;
		}
		return 0;
	case ARRAY_QUADRUPLE:
		if (type_src == ARRAY_QUADRUPLE) {
			*(reinterpret_cast<zeQuadruple*>(dst)) = *(reinterpret_cast<const zeQuadruple*>(src));
			return 1;
		}
		if (type_src == ARRAY_DOUBLE) {
			const zeDouble *s = reinterpret_cast<const zeDouble*>(src);
			zeQuadruple *d = reinterpret_cast<zeQuadruple*>(dst);
			if (s->ndat() < 4) return 0;
			d->resize(s->ndat() / 4, 1);
			memcpy(d->data(), s->data(), d->ndat()*sizeof(sQuadruple));
			return 1;
		}
		return 0;
	}
	return 0;
}

int
array_import(void *dst, const void *src, size_t n, size_t esize)
{
	int type = array_check_type(dst);
	if (!type) return 0;
	array_set_error_message("array_import() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		if (esize != 1) return 0;
		{
			zeChar* arr = reinterpret_cast<zeChar*>(dst);
			arr->resize(n, 1);
			memcpy(arr->data(), src, n * esize);
			return 1;
		}
		return 0;
	case ARRAY_UCHAR:
		if (esize != 1) return 0;
		{
			zeUchar* arr = reinterpret_cast<zeUchar*>(dst);
			arr->resize(n, 1);
			memcpy(arr->data(), src, n * esize);
			return 1;
		}
		return 0;
	case ARRAY_SHORT:
		if (esize != sizeof(short)) return 0;
		{
			zeShort* arr = reinterpret_cast<zeShort*>(dst);
			arr->resize(n, 1);
			memcpy(arr->data(), src, n * esize);
			return 1;
		}
		return 0;
	case ARRAY_USHORT:
		if (esize != sizeof(unsigned short)) return 0;
		{
			zeUshort* arr = reinterpret_cast<zeUshort*>(dst);
			arr->resize(n, 1);
			memcpy(arr->data(), src, n * esize);
			return 1;
		}
		return 0;
	case ARRAY_INT:
		if (esize != sizeof(int)) return 0;
		{
			zeInt* arr = reinterpret_cast<zeInt*>(dst);
			arr->resize(n, 1);
			memcpy(arr->data(), src, n * esize);
			return 1;
		}
		return 0;
	case ARRAY_UINT:
		if (esize != sizeof(unsigned int)) return 0;
		{
			zeUint* arr = reinterpret_cast<zeUint*>(dst);
			arr->resize(n, 1);
			memcpy(arr->data(), src, n * esize);
			return 1;
		}
		return 0;
	case ARRAY_FLOAT:
		if (esize != sizeof(float)) return 0;
		{
			zeFloat* arr = reinterpret_cast<zeFloat*>(dst);
			arr->resize(n, 1);
			memcpy(arr->data(), src, n * esize);
			return 1;
		}
		return 0;
	case ARRAY_DOUBLE:
		if (esize != sizeof(double)) return 0;
		{
			zeDouble* arr = reinterpret_cast<zeDouble*>(dst);
			arr->resize(n, 1);
			memcpy(arr->data(), src, n * esize);
			return 1;
		}
		return 0;
	case ARRAY_COMPLEX:
		if (esize != sizeof(sComplex)) return 0;
		{
			zeComplex* arr = reinterpret_cast<zeComplex*>(dst);
			arr->resize(n, 1);
			memcpy(arr->data(), src, n * esize);
			return 1;
		}
		return 0;
	case ARRAY_TRIPLEX:
		if (esize != sizeof(sTriplex)) return 0;
		{
			zeTriplex* arr = reinterpret_cast<zeTriplex*>(dst);
			arr->resize(n, 1);
			memcpy(arr->data(), src, n * esize);
			return 1;
		}
		return 0;
	case ARRAY_QUADRUPLE:
		if (esize != sizeof(sQuadruple)) return 0;
		{
			zeQuadruple* arr = reinterpret_cast<zeQuadruple*>(dst);
			arr->resize(n, 1);
			memcpy(arr->data(), src, n * esize);
			return 1;
		}
		return 0;
	}
	return 1;
}

int
array_parse(void *ptr, const char* src, size_t i)
{
	int type = array_check_type(ptr);
	if (!type) return 0;
	array_set_error_message("array_parse() error.");
	switch (type)
	{
	case ARRAY_CHAR:
		return reinterpret_cast<zeChar*>(ptr)->parse(i, src);
	case ARRAY_UCHAR:
		return reinterpret_cast<zeUchar*>(ptr)->parse(i, src);
	case ARRAY_SHORT:
		return reinterpret_cast<zeShort*>(ptr)->parse(i, src);
	case ARRAY_USHORT:
		return reinterpret_cast<zeUshort*>(ptr)->parse(i, src);
	case ARRAY_INT:
		return reinterpret_cast<zeInt*>(ptr)->parse(i, src);
	case ARRAY_UINT:
		return reinterpret_cast<zeUint*>(ptr)->parse(i, src);
	case ARRAY_FLOAT:
		return reinterpret_cast<zeFloat*>(ptr)->parse(i, src);
	case ARRAY_DOUBLE:
		return reinterpret_cast<zeDouble*>(ptr)->parse(i, src);
	}
	return 0;
}

bool has_only_numbers(char *s, bool csv)
{
	int i, len = strlen(s);
	for (i = 0; s[i]; i++) {
		if (isalnum(s[i]) || ispunct(s[i])) break;
	}
	if (!s[i]) return false;		// empty line
	i = 0;
	while (s[i]) {
		if (csv && s[i] == ',') {
			s[i] = ' ';
			i++;
			continue;
		}
		if (isdigit(s[i]) || isspace(s[i])) {
			i++;
			continue;
		}
		switch (s[i]) {
		case '-':case '+':
			if (isdigit(s[i+1]) || s[i+1] == '.' || s[i+1] == 'e' || s[i+1] == 'E') break;
			return false;
		case 'e':case 'E':
			if (isdigit(s[i+1])) break;
			if (s[i+1] == '-' || s[i+1] == '+') {
				if (i+2 < len && isdigit(s[i+2])) break;
			}
			return false;
		case '.':
			if (isdigit(s[i+1]) || s[i+1] == 'e' || s[i+1] == 'E') break;
			return false;
		default:
			return false;
		}
		i++;
	}
	return true;
}

int
array_read_text(void *ptr, const char *filename)
{
	if (!array_is_double(ptr)) return 0;

	FILE *f = fopen(filename, "r");
	if (!f) {
		array_set_error_message2(filename, ": cannot open the file for reading.");
		return 0;
	}
	
	int nrow = 0, ncol = 0, max_len = 65536;
	zeChar obj;
	obj.resize(max_len+1, 1);
	char *buf = obj.data();

	while (!feof(f)) {
		if (!fgets(buf, max_len, f)) continue;
		if (strlen(buf) == max_len) {
			array_set_error_message2(filename, ": lines too long.");
			return 0;
		}
		if (has_only_numbers(buf, true)) {
			nrow++;
			if (ncol == 0) {
				int i = 0, i0 = 0;
				while (buf[i]) {
					if (isspace(buf[i]) || !buf[i+1]) {
						if (i0 != i) ncol++;
						i++;
						i0 = i;
					}
					else {
						i++;
					}
				}
			}
		}
	}

	if (nrow == 0 || ncol == 0) {
		array_set_error_message2(filename, ": no line contains only numbers.");
		return 0;
	}

	zeDouble *arr = (zeDouble*)ptr;
	arr->resize(nrow, ncol);
	fseek(f, 0, SEEK_SET);
	nrow = 0;

	while (!feof(f)) {
		fgets(buf, max_len, f);
		if (has_only_numbers(buf, true)) {
			arr->parse(nrow, buf);
			nrow++;
		}
	}

	fclose(f);

	return 1;
}