#ifndef __ZE_ARRAY_H__
#define __ZE_ARRAY_H__

#ifdef ZE_EXPORTS
	#define ZE_API __declspec(dllexport)
#elif ZE_IMPORTS 
	#define ZE_API __declspec(dllimport)
#else
	#define ZE_API
#endif

#include <stdio.h>

#ifdef __cplusplus
extern "C" {
#endif

ZE_API
void array_set_error_message(const char* msg);

ZE_API
void array_set_error_message2(const char* msg1, const char* msg2);

ZE_API
const char* array_get_error_message();

ZE_API
const char* array_get_typename(const void *ptr);

ZE_API
int array_check_type(const void *ptr);

ZE_API
int array_check_type2(const void *ptr1, const void *ptr2);

ZE_API
int array_is_int8(const void *ptr);

ZE_API
int array_is_uint8(const void *ptr);

ZE_API
int array_is_int16(const void *ptr);

ZE_API
int array_is_uint16(const void *ptr);

ZE_API
int array_is_int32(const void *ptr);

ZE_API
int array_is_uint32(const void *ptr);

ZE_API
int array_is_float(const void *ptr);

ZE_API
int array_is_double(const void *ptr);

ZE_API
int array_is_complex(const void *ptr);

ZE_API
int array_is_triplex(const void *ptr);

ZE_API
int array_is_quadruple(const void *ptr);

ZE_API
void *array_create(const char *type, size_t nrow, size_t ncol);

ZE_API
int array_free(void *ptr);

ZE_API
void * array_get_ptr(void * ptr, size_t *esize, size_t *n);

ZE_API
int array_get_size(const void *ptr, size_t *nrow, size_t *ncol);

ZE_API
int array_resize(void *ptr, size_t nrow, size_t ncol, bool keep);

ZE_API
int array_reshape(void *ptr, size_t nrow, size_t ncol);

ZE_API
int array_fill(void *ptr, double v, double dv);

ZE_API
int array_negate(void *ptr);

ZE_API
int array_add(void *ptr, double v);

ZE_API
int array_add2(void *dst, const void *src);

ZE_API
int array_subtract(void *ptr, double v);

ZE_API
int array_subtract2(void *dst, const void *src);

ZE_API
int array_multiply(void *ptr, double v);

ZE_API
int array_multiply2(void *dst, const void *src);

ZE_API
int array_element_inverse(void *ptr);

ZE_API
int array_divide(void *ptr, double v);

ZE_API
int array_divide2(void *dst, const void *src);

ZE_API
int array_and(void *idx, const void *ptr1, const void *ptr2);

ZE_API
int array_or(void *idx, const void *ptr1, const void *ptr2);

ZE_API
int array_xor(void *idx, const void *ptr1, const void *ptr2);

ZE_API
int array_equal(void *idx, const void *ptr, double v);

ZE_API
int array_equal2(void *idx, const void *ptr1, const void *ptr2);

ZE_API
int array_nequal(void *idx, const void *ptr, double v);

ZE_API
int array_nequal2(void *idx, const void *ptr1, const void *ptr2);

ZE_API
int array_less(void *idx, const void *ptr, double v);

ZE_API
int array_less2(void *idx, const void *ptr1, const void *ptr2);

ZE_API
int array_lequal(void *idx, const void *ptr, double v);

ZE_API
int array_lequal2(void *idx, const void *ptr1, const void *ptr2);

ZE_API
int array_greater(void *idx, const void *ptr, double v);

ZE_API
int array_greater2(void *idx, const void *ptr1, const void *ptr2);

ZE_API
int array_gequal(void *idx, const void *ptr, double v);

ZE_API
int array_gequal2(void *idx, const void *ptr1, const void *ptr2);

ZE_API
double array_get_at(const void *ptr, size_t i, size_t j);

ZE_API
int array_set_at(void *ptr, size_t i, size_t j, double v);

ZE_API
int array_get_row(void *dst, const void *src, size_t j);

ZE_API
int array_get_column(void *dst, const void *src, size_t j);

ZE_API
int array_set_row(void *dst, double v, size_t i);

ZE_API
int array_set_row2(void *dst, const void *src, size_t i);

ZE_API
int array_set_column(void *dst, double v, size_t i);

ZE_API
int array_set_column2(void *dst, const void *src, size_t i);

ZE_API
int array_insert_row(void *dst, double v, size_t i);

ZE_API
int array_insert_row2(void *dst, const void* src, size_t i);

ZE_API
int array_insert_column(void *dst, double v, size_t i);

ZE_API
int array_insert_column2(void *dst, const void* src, size_t i);

ZE_API
int array_delete_row(void *ptr, size_t i);

ZE_API
int array_delete_column(void *ptr, size_t i);

ZE_API
int array_get_by_index(void *dst, const void *src, const void *idx);

ZE_API
int array_set_by_index(void *dst, double v, const void *idx);

ZE_API
int array_set_by_index2(void *dst, const void *src, const void *idx);

ZE_API
int array_get_by_block(void *dst, const void *src, size_t i, size_t m, size_t j, size_t n);

ZE_API
int array_set_by_block(void *dst, const void *src, size_t i, size_t j);

ZE_API
int array_find_eq(void *idx, const void *ptr, double v);

ZE_API
int array_find_ne(void *idx, const void *ptr, double v);

ZE_API
int array_transpose(void *ptr);

ZE_API
int array_swap_rows(void *ptr, size_t i, size_t j);

ZE_API
int array_swap_columns(void *ptr, size_t i, size_t j);

ZE_API
int array_flip_rows(void *ptr);

ZE_API
int array_flip_columns(void *ptr);

ZE_API
int array_shift_rows(void *ptr, int n);

ZE_API
int array_shift_columns(void *ptr, int n);

ZE_API
int array_product(void *dst, const void *src);

ZE_API
int array_sort(void *ptr, size_t i, int assending);

ZE_API
int array_unique(void *ptr, int assending);

ZE_API
double array_min(const void *ptr);

ZE_API
double array_max(const void *ptr);

ZE_API
double array_sum(const void *ptr);

ZE_API
int array_confine(void *ptr, double vmin, double vmax);

ZE_API
int array_print(const void *ptr, const char *format, FILE *f = stdout);

ZE_API
int array_clone(void *dst, const void *src);

ZE_API
int array_import(void *dst, const void *src, size_t n, size_t esize);

ZE_API
int array_parse(void *ptr, const char* src, size_t i);

ZE_API
int array_read_text(void *ptr, const char *filename);


#ifdef __cplusplus
}
#endif 

#endif